import streamlit as st
import streamlit_antd_components as sac
import numpy as np
from plotly import graph_objects as go
from antelope.connection import connect, import_schemas
from plotly.subplots import make_subplots
import plotly.express as px
import pandas as pd
import matplotlib.pyplot as plt
import probeinterface as pi
from probeinterface.plotting import plot_probe_group


@st.cache_data(ttl=3600)
def session_length(session_dict):

    # open database connection for main thread
    conn = connect.connect()

    # import schema structure
    tables = import_schemas.schema(conn)

    # loop through all the data
    max_time = 0
    units = (tables["SpikeTrain"] & session_dict).fetch(as_dict=True)
    for unit in units:
        max_time = max(max_time, np.max(unit["spiketrain"], initial=0))
    analog_events = (tables["AnalogEvents"] & session_dict).fetch(as_dict=True)
    for event in analog_events:
        max_time = max(max_time, np.max(event["timestamps"], initial=0))
    digital_events = (tables["DigitalEvents"] & session_dict).fetch(as_dict=True)
    for event in digital_events:
        max_time = max(max_time, np.max(event["timestamps"], initial=0))
    interval_events = (tables["IntervalEvents"] & session_dict).fetch(as_dict=True)
    for event in interval_events:
        max_time = max(max_time, np.max(event["timestamps"], initial=0))

    return max_time


@st.cache_data(ttl=3600, max_entries=3)
def plot_session(session_dict, selected_animals, selected_objects, timerange):

    # unpack objects
    (
        analog_groups,
        digital_groups,
        interval_objects,
        kinematics_objects,
        masks_objects,
    ) = selected_objects

    # open database connection for main thread
    conn = connect.connect()

    # import schema structure
    tables = import_schemas.schema(conn)

    # get spike trains for the selected animals and sorting parameters - requires a slight abuse of union
    query = False
    for i, (animal_id, param_id, selected_units, selected_lfps) in enumerate(
        selected_animals
    ):

        if i == 0:
            query = (
                tables["SpikeTrain"].proj()
                & session_dict
                & {"animal_id": animal_id, "sortingparams_id": param_id}
                & selected_units
            )
        else:
            query = query + (
                tables["SpikeTrain"].proj()
                & session_dict
                & {"animal_id": animal_id, "sortingparams_id": param_id}
                & selected_units
            )

    spiketrains = (
        tables["SpikeTrain"] * tables["Animal"].proj("animal_name") & query
    ).fetch()

    # get lfps for the selected animals and sorting parameters
    query = False
    for i, (animal_id, param_id, selected_units, selected_lfps) in enumerate(
        selected_animals
    ):

        if i == 0:
            query = (
                tables["LFP"].proj()
                & session_dict
                & {"animal_id": animal_id, "sortingparams_id": param_id}
                & selected_lfps
            )
        else:
            query = query + (
                tables["LFP"].proj()
                & session_dict
                & {"animal_id": animal_id, "sortingparams_id": param_id}
                & selected_lfps
            )

    lfps = (tables["LFP"] * tables["Animal"].proj("animal_name") & query).fetch()

    # update spiketrains to only include the selected time range
    for i, unit in enumerate(spiketrains):
        spiketrain = unit["spiketrain"]
        spiketrains[i]["spiketrain"] = spiketrain[
            (spiketrain >= timerange[0]) & (spiketrain <= timerange[1])
        ]

    # update lfps to only include the selected time range
    for i, lfp in enumerate(lfps):
        sample_rate = lfp["lfp_sample_rate"]
        start_sample = int(timerange[0] * sample_rate)
        end_sample = int(timerange[1] * sample_rate)
        lfps[i]["lfp"] = lfp["lfp"][start_sample:end_sample]

    # calculate animals
    animals = np.unique(spiketrains["animal_name"])

    # calculate unit numbers per animal
    unit_numbers = [
        len(spiketrains[spiketrains["animal_name"] == animal]) for animal in animals
    ]

    # make list of subplot heights
    subplot_heights = [30 + i * 6 for i in unit_numbers]
    subplot_heights += [20] * len(lfps)
    subplot_heights += [50] * len(analog_groups)
    subplot_heights += [50] * len(digital_groups)
    if len(interval_objects) > 0:
        subplot_heights.append(8 * len(interval_objects))
    if len(kinematics_objects) > 0:
        subplot_heights.append(8 * len(kinematics_objects))
        subplot_heights.append(8 * len(kinematics_objects))
    if len(masks_objects) > 0:
        subplot_heights.append(8 * len(masks_objects))

    # create figure
    fig = make_subplots(
        rows=len(subplot_heights),
        cols=1,
        shared_xaxes=True,
        vertical_spacing=0.1,
        row_heights=subplot_heights,
    )

    # loop through animals
    for i, animal in enumerate(animals):

        # calculate probes for each animal
        probes = np.unique(
            spiketrains[spiketrains["animal_name"] == animal]["probe_id"]
        )

        # keep track of y position and unit labels
        y_pos = 1
        y_text = []

        # loop through probes
        for j, probe in enumerate(probes):

            # calculate units for each animal and probe
            units = np.unique(
                spiketrains[
                    (spiketrains["animal_name"] == animal)
                    & (spiketrains["probe_id"] == probe)
                ]["unit_id"]
            )

            # loop through units
            for k, unit in enumerate(units):

                # store y label
                y_text.append(str(probe) + "-" + str(unit))

                # get spike times for each unit
                spiketrain = spiketrains[
                    (spiketrains["animal_name"] == animal)
                    & (spiketrains["probe_id"] == probe)
                    & (spiketrains["unit_id"] == unit)
                ]["spiketrain"][0]
                spiketrain = spiketrain.astype("float32")

                # create y values for each unit
                y = np.ones_like(spiketrain, dtype="uint8") * y_pos

                # add unit to plot
                fig.add_trace(
                    go.Scattergl(
                        x=spiketrain,
                        y=y,
                        mode="markers",
                        marker=dict(symbol="line-ns-open"),
                        name=str(unit),
                        showlegend=False,
                    ),
                    row=i + 1,
                    col=1,
                )

                # update y position
                y_pos += 1

        # update axes
        fig.update_yaxes(
            title_text=f"{animal} spiketrains",
            tickvals=np.arange(1, y_pos),
            ticktext=y_text,
            row=i + 1,
            col=1,
        )

    # current row
    row = len(animals)

    for i, lfp in enumerate(lfps):

        # make timestamps
        timestamps = np.arange(lfp["lfp"].size) / lfp["lfp_sample_rate"] + timerange[0]

        name = f"{lfp['probe_id']}-{lfp['channel_id']}"

        # add lfp to plot
        fig.add_trace(
            go.Scattergl(
                x=timestamps, y=lfp["lfp"], mode="lines", name=name, showlegend=False
            ),
            row=row + 1,
            col=1,
        )
        fig.update_yaxes(title_text=name, row=row + 1, col=1)
        row += 1

    # loop through analog events
    for i, (name, group) in enumerate(analog_groups):
        row += 1

        # pull data
        groupdata = (
            tables["AnalogEvents"] * tables["Object"].proj("object_name")
            & session_dict
            & group
        ).fetch(as_dict=True)

        # filter data to be in timerange
        for j, event in enumerate(groupdata):
            timestamps = event["timestamps"]
            data = event["data"]
            groupdata[j]["timestamps"] = timestamps[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]
            groupdata[j]["data"] = data[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]

        # plot events
        for event in groupdata:
            if event["timestamps"].size > 0:
                fig.add_trace(
                    go.Scattergl(
                        x=event["timestamps"],
                        y=event["data"],
                        mode="lines",
                        name=event["object_name"],
                        showlegend=True,
                    ),
                    row=row,
                    col=1,
                )

        fig.update_yaxes(title_text=f"{name} ({event['unit']})", row=row, col=1)

    # loop through digital events
    for i, (name, group) in enumerate(digital_groups):
        row += 1

        # pull data
        groupdata = (
            tables["DigitalEvents"] * tables["Object"].proj("object_name")
            & session_dict
            & group
        ).fetch(as_dict=True)

        # filter data to be in timerange
        for j, event in enumerate(groupdata):
            timestamps = event["timestamps"]
            data = event["data"]
            groupdata[j]["timestamps"] = timestamps[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]
            groupdata[j]["data"] = data[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]

        # plot events
        for event in groupdata:
            fig.add_trace(
                go.Scattergl(
                    x=event["timestamps"],
                    y=event["data"],
                    mode="markers",
                    name=event["object_name"],
                    showlegend=True,
                ),
                row=row,
                col=1,
            )

        fig.update_yaxes(title_text=f"{name} ({event['unit']})", row=row, col=1)

    # loop through interval events
    if len(interval_objects) > 0:
        row += 1
        y_text = []

        # pull data
        interval_events = (
            tables["IntervalEvents"] * tables["Object"].proj("object_name")
            & session_dict
            & interval_objects
        ).fetch(as_dict=True)

        # filter data to be in timerange
        for j, event in enumerate(interval_events):
            timestamps = event["timestamps"]
            data = event["data"]
            interval_events[j]["timestamps"] = timestamps[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]
            interval_events[j]["data"] = data[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]

        for i, event in enumerate(interval_events):

            y_text.append(event["object_name"])

            # get timestamps and data
            timestamps = event["timestamps"]
            data = event["data"]

            # clean data
            if len(data) > 0:
                if data[0] == -1:
                    data = data[1:]
                    timestamps = timestamps[1:]
            if len(data) > 0:
                if data[-1] == 1:
                    data = data[:-1]
                    timestamps = timestamps[:-1]

            # slightly hacky - convert -1 to 2, then duplicate timestamps
            data = (-data + 3) // 2
            timestamps = np.repeat(timestamps, data)

            # now make y values, where every third value is none
            y = np.ones_like(timestamps) * i
            y[2::3] = None

            fig.add_trace(
                go.Scattergl(
                    x=timestamps,
                    y=y,
                    mode="lines",
                    name=event["object_name"],
                    showlegend=False,
                ),
                row=row,
                col=1,
            )

        fig.update_yaxes(
            title_text="Interval events",
            tickvals=np.arange(len(y_text)),
            ticktext=y_text,
            row=row,
            col=1,
        )

    # loop through kinematics events
    if len(kinematics_objects) > 0:

        row += 1

        # pull data
        kinematics_events = (
            tables["Kinematics"] * tables["Object"].proj("object_name")
            & session_dict
            & kinematics_objects
        ).fetch(as_dict=True)

        # filter data to be in timerange
        for j, event in enumerate(kinematics_events):

            timestamps = event["timestamps"]
            x, y = event["data"][:, 0], event["data"][:, 1]
            kinematics_events[j]["timestamps"] = timestamps[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]
            kinematics_events[j]["x"] = x[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]
            kinematics_events[j]["y"] = y[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]

            # get timestamps and data
            timestamps = event["timestamps"]
            x = event["x"]
            y = event["y"]

            # plot events
            if event["timestamps"].size > 0:
                fig.add_trace(
                    go.Scattergl(
                        x=timestamps,
                        y=x,
                        mode="lines",
                        name=f"""{event['object_name']} x-axis""",
                        showlegend=True,
                    ),
                    row=row,
                    col=1,
                )
                fig.add_trace(
                    go.Scattergl(
                        x=timestamps,
                        y=y,
                        mode="lines",
                        name=f"""{event['object_name']} y-axis""",
                        showlegend=True,
                    ),
                    row=row + 1,
                    col=1,
                )

        fig.update_yaxes(title_text="Kinematics x-axis", row=row, col=1)
        fig.update_yaxes(title_text="Kinematics y-axis", row=row + 1, col=1)

        row += 1

    # loop through masks events
    if len(masks_objects) > 0:
        row += 1
        y_text = []

        # pull data
        masks_events = (tables["Mask"] & session_dict & masks_objects).fetch(
            as_dict=True
        )

        # filter data to be in timerange
        for j, event in enumerate(masks_events):
            timestamps = event["timestamps"]
            data = event["data"]
            masks_events[j]["timestamps"] = timestamps[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]
            masks_events[j]["data"] = data[
                (timestamps >= timerange[0]) & (timestamps <= timerange[1])
            ]

        for i, event in enumerate(masks_events):

            y_text.append(event["mask_name"])

            # get timestamps and data
            timestamps = event["timestamps"]
            data = event["data"]

            # clean data
            if len(data) > 0:
                if data[0] == -1:
                    data = data[1:]
                    timestamps = timestamps[1:]
            if len(data) > 0:
                if data[-1] == 1:
                    data = data[:-1]
                    timestamps = timestamps[:-1]

            # slightly hacky - convert -1 to 2, then duplicate timestamps
            data = (-data + 3) // 2
            data = data.astype("int", casting="unsafe")
            timestamps = np.repeat(timestamps, data)

            # now make y values, where every third value is none
            y = np.ones_like(timestamps) * i
            y[2::3] = None

            fig.add_trace(
                go.Scattergl(
                    x=timestamps,
                    y=y,
                    mode="lines",
                    name=event["mask_name"],
                    showlegend=False,
                ),
                row=row,
                col=1,
            )

        fig.update_yaxes(
            title_text="Masks",
            tickvals=np.arange(len(y_text)),
            ticktext=y_text,
            row=row,
            col=1,
        )

    # custom layout config that uses the last subplot in a hacky way
    row = "" if row == 1 else row
    fig["layout"][f"xaxis{row}_rangeslider_visible"] = True
    fig["layout"][f"xaxis{row}_rangeslider_thickness"] = 0.05
    fig["layout"][f"xaxis{row}_title"] = "Time (seconds)"
    fig["layout"][f"xaxis{row}_range"] = timerange

    # global layout config
    fig.update_layout(height=sum(subplot_heights) * 5, width=800)

    return fig


@st.cache_data(ttl=3600, max_entries=3)
def plot_kinematics(kinematics_dict, timerange):

    # open database connection for main thread
    conn = connect.connect()

    # import schema structure
    tables = import_schemas.schema(conn)

    # pull data
    kinematics_events = (
        tables["Kinematics"] * tables["Object"].proj("object_name") & kinematics_dict
    ).fetch(as_dict=True)

    # filter data to be in timerange
    for j, event in enumerate(kinematics_events):
        timestamps = event["timestamps"]
        data = event["data"]
        kinematics_events[j]["timestamps"] = timestamps[
            (timestamps >= timerange[0]) & (timestamps <= timerange[1])
        ]
        kinematics_events[j]["data"] = data[
            (timestamps >= timerange[0]) & (timestamps <= timerange[1])
        ]

    for i, event in enumerate(kinematics_events):

        df = pd.DataFrame(
            {
                "time": kinematics_events[i]["timestamps"],
                "x": kinematics_events[i]["data"][:, 0],
                "y": kinematics_events[i]["data"][:, 1],
            }
        )

        date_to_val = df["time"].map(
            pd.Series(data=np.arange(len(df)), index=df["time"].values).to_dict()
        )
        timerange = df["time"].max() - df["time"].min()
        tickvals = [int(timerange) * k * 2 for k in range(1, 15)]
        dlist = list(date_to_val)
        index_tickvals = [dlist.index(tv) for tv in tickvals]
        ticktext = [format(df["time"][id], ".1f") for id in index_tickvals]

        kin_fig = go.Figure(
            go.Scattergl(
                x=df["x"],
                y=df["y"],
                mode="markers",
                marker_color=date_to_val,
                marker_colorscale="Plasma",
                marker_showscale=True,
                marker_size=8,
                marker_colorbar=dict(
                    tickvals=tickvals, ticktext=ticktext, title_text="Time (s)"
                ),
                customdata=df["time"],
                hovertemplate="%{customdata}<br>x: %{x}<br>y: %{y}",
            )
        )

        kin_fig.update_layout(
            width=800,
            height=700,
            title_text=f'{event["object_name"].capitalize()} kinematics',
            xaxis_title="x (pixels)",
            yaxis_title="y (pixels)",
        )

    return kin_fig


def choose_animals(tables, session_dict):

    # calculate what animals have spikesorted data
    animals = (
        (
            tables["Animal"] * tables["SpikeSorting"] * tables["SortingParams"]
            & session_dict
        )
        .proj("animal_name", "sortingparams_name")
        .fetch()
    )

    # loop through animals, and make dict of animal ids to animal names
    animal_dict = {}
    for animal in animals:
        animal_dict[animal["animal_name"]] = animal["animal_id"]

    # make another dict that keeps track of the sorting parameters for each animal
    params_dict = {}
    for animal_id in animal_dict.values():
        ids = animals[animals["animal_id"] == animal_id]["sortingparams_id"]
        names = animals[animals["animal_id"] == animal_id]["sortingparams_name"]
        params_dict[animal_id] = {name: id for id, name in zip(ids, names)}

    # get user to select the animals they want to visualise
    if len(animal_dict) > 1:
        selected_animals = st.multiselect(
            "Select the animals you want to visualise",
            list(animal_dict.keys()),
            list(animal_dict.keys())[0],
        )
    elif len(animal_dict) == 1:
        selected_animals = list(animal_dict.keys())
    else:
        return []

    st.divider()
    st.markdown("**Neural activity**")

    # loop through selected animals, and choose a unique sorting parameter if there are multiple
    animals_with_sorting = []
    for animal in selected_animals:
        if len(params_dict[animal_dict[animal]]) > 1:
            param_id = params_dict[animal_dict[animal]][
                st.selectbox(
                    f"The animal '{animal}' has been spikesorted multiple times, please select which sorting parameter set you want to visualise",
                    list(params_dict[animal_dict[animal]].keys()),
                )
            ]
        else:
            param_id = list(params_dict[animal_dict[animal]].values())[0]

        # now, for this animal, select the units you want to visualise
        # first query the probes
        probes = (
            tables["Probe"]
            & session_dict
            & {"animal_id": animal_dict[animal], "sortingparams_id": param_id}
        ).fetch(as_dict=True)

        # initialise selected units
        selected_units = []
        selected_lfps = []

        st.text("")
        st.markdown("***Units***")
        st.text(f'Select the units you want to visualise for animal "{animal}"')

        # loop through probes and get user to select units
        for probe in probes:

            # get units for this probe
            units = (
                tables["Unit"]
                & session_dict
                & {
                    "animal_id": animal_dict[animal],
                    "sortingparams_id": param_id,
                    "probe_id": probe["probe_id"],
                }
            ).fetch(as_dict=True)

            # make a dict mapping unit names to unit dicts
            unit_dict = {}
            for unit in units:
                unit_dict[str(unit["unit_id"])] = unit

            # get user to select the units they want to visualise
            selected_units += [
                unit_dict[i]
                for i in sac.checkbox(
                    items=list(unit_dict.keys()),
                    label=f"""Probe {probe['probe_id']} units""",
                    index=list(range(len(unit_dict))),
                    check_all="Select all",
                )
            ]

        st.markdown("***LFPs***")
        st.text(f'Select the LFPs you want to visualise for animal "{animal}"')

        # loop through probes and get user to select units
        for probe in probes:

            # get units for this probe
            lfps = (
                tables["Channel"].proj()
                & session_dict
                & {
                    "animal_id": animal_dict[animal],
                    "sortingparams_id": param_id,
                    "probe_id": probe["probe_id"],
                }
            ).fetch(as_dict=True)

            # make a dict mapping unit names to unit dicts
            lfp_dict = {}
            for lfp in lfps:
                lfp_dict[str(lfp["channel_id"])] = lfp

            # get user to select the units they want to visualise
            selected_lfps += [
                lfp_dict[i]
                for i in sac.checkbox(
                    items=list(lfp_dict.keys()),
                    label=f"""Probe {probe['probe_id']} channels""",
                    check_all="Select all",
                )
            ]

        animals_with_sorting.append(
            (animal_dict[animal], param_id, selected_units, selected_lfps)
        )

    return animals_with_sorting


def choose_kinematics(tables, session_dict):

    if len(tables["Kinematics"] & session_dict) == 0:
        return []
    st.divider()

    # get the objects for this session
    kinematics = (tables["Object"] * tables["Kinematics"].proj() & session_dict).fetch(
        as_dict=True
    )

    # for interval events, just select any objects
    st.markdown("***Kinematics object***")
    kinematics_dict = {obj["object_name"]: obj for obj in kinematics}
    selected_object = kinematics_dict[
        st.selectbox("Select kinematics object", list(kinematics_dict.keys()))
    ]

    return selected_object


def choose_objects(tables, session_dict):

    if (
        len(tables["AnalogEvents"] & session_dict) == 0
        and len(tables["DigitalEvents"] & session_dict) == 0
        and len(tables["IntervalEvents"] & session_dict) == 0
        and len(tables["Kinematics"] & session_dict) == 0
        and len(tables["Mask"] & session_dict) == 0
    ):
        return ([], [], [], [], [])
    st.divider()
    st.markdown("**Environment features**")

    # get the objects for this session
    analog = (
        tables["Object"] * tables["AnalogEvents"].proj("unit") & session_dict
    ).fetch(as_dict=True)
    digital = (
        tables["Object"] * tables["DigitalEvents"].proj("unit") & session_dict
    ).fetch(as_dict=True)
    interval = (
        tables["Object"] * tables["IntervalEvents"].proj() & session_dict
    ).fetch(as_dict=True)
    kinematics = (tables["Object"] * tables["Kinematics"].proj() & session_dict).fetch(
        as_dict=True
    )
    masks = (tables["Mask"].proj("mask_name") & session_dict).fetch(as_dict=True)

    # for analog events, we want the user to be able to add a number of groups with a common y axis
    analog_groups = []
    analog_used = []
    st.markdown("***Analog events***")
    i = 1
    keep_going = st.checkbox(f"Add group {i}", key=f"Add group {i}")
    while keep_going:

        # get the user to select objects not already in a group
        analog_dict = {
            obj["object_name"]: obj
            for obj in analog
            if obj["object_name"] not in analog_used
        }
        name = st.text_input("Enter name", list(analog_dict.keys())[0])
        selected_objects = sac.checkbox(
            items=analog_dict.keys(),
            label=f"Select the objects you want to visualise for group {i}",
            check_all="Select all",
        )

        # ask user if they want a new group
        i += 1
        if len(selected_objects) > 0:

            # check all selected objects have the same unit
            if (
                len(
                    set(
                        [analog_dict[obj_name]["unit"] for obj_name in selected_objects]
                    )
                )
                > 1
            ):
                st.error("All objects in a group must have the same unit")
                keep_going = False

            else:

                analog_used += selected_objects
                analog_groups.append(
                    (name, [analog_dict[obj_name] for obj_name in selected_objects])
                )

                if (
                    len(
                        [obj for obj in analog if obj["object_name"] not in analog_used]
                    )
                    > 0
                ):
                    keep_going = st.checkbox(f"Add group {i}", key=f"Add group {i}")

                else:
                    keep_going = False

        else:
            keep_going = False

    # for digital events, we want the user to be able to add a number of groups with a common y axis
    digital_groups = []
    digital_used = []
    st.markdown("***Digital events***")
    i = 1
    keep_going = st.checkbox(f"Add group {i}", key=f"Add digital group {i}")
    while keep_going:

        # get the user to select objects not already in a group
        digital_dict = {
            obj["object_name"]: obj
            for obj in digital
            if obj["object_name"] not in digital_used
        }
        name = st.text_input("Enter name", list(digital_dict.keys())[0])
        selected_objects = sac.checkbox(
            items=digital_dict.keys(),
            label=f"Select the objects you want to visualise for group {i}",
            check_all="Select all",
        )

        # ask user if they want a new group
        i += 1
        if len(selected_objects) > 0:

            # check all selected objects have the same unit
            if (
                len(
                    set(
                        [
                            digital_dict[obj_name]["unit"]
                            for obj_name in selected_objects
                        ]
                    )
                )
                > 1
            ):
                st.error("All objects in a group must have the same unit")
                keep_going = False

            else:

                digital_used += selected_objects
                digital_groups.append(
                    (name, [digital_dict[obj_name] for obj_name in selected_objects])
                )

                if (
                    len(
                        [
                            obj
                            for obj in digital
                            if obj["object_name"] not in digital_used
                        ]
                    )
                    > 0
                ):
                    keep_going = st.checkbox(
                        f"Add group {i}", key=f"Add digital group {i}"
                    )

                else:
                    keep_going = False

        else:
            keep_going = False

    # for interval events, just select any objects
    st.markdown("***Interval events***")
    interval_dict = {obj["object_name"]: obj for obj in interval}
    selected_objects = sac.checkbox(
        items=interval_dict.keys(),
        label="Select the interval events you want to visualise",
        check_all="Select all",
    )

    interval_objects = [interval_dict[obj_name] for obj_name in selected_objects]

    # for interval events, just select any objects
    st.markdown("***Kinematics object***")
    kinematics_dict = {obj["object_name"]: obj for obj in kinematics}
    selected_objects = sac.checkbox(
        items=kinematics_dict.keys(),
        label="Select the kinematics objects you want to visualise",
        check_all="Select all",
    )

    kinematics_objects = [kinematics_dict[obj_name] for obj_name in selected_objects]

    # for mask events, just select any objects
    st.markdown("***Masks***")
    mask_dict = {obj["mask_name"]: obj for obj in masks}
    selected_objects = sac.checkbox(
        items=mask_dict.keys(),
        label="Select the mask events you want to visualise",
        check_all="Select all",
    )

    mask_objects = [mask_dict[obj_name] for obj_name in selected_objects]

    return (
        analog_groups,
        digital_groups,
        interval_objects,
        kinematics_objects,
        mask_objects,
    )


@st.cache_data(ttl=3600, max_entries=3)
def plot_units(spikesorting_dict, selected_units):

    # open database connection for main thread
    conn = connect.connect()

    # import schema structure
    tables = import_schemas.schema(conn)

    # fetch waveform data
    waveforms = (tables["Waveform"] & spikesorting_dict & selected_units).fetch()

    # loop through probes
    probes = np.unique(waveforms["probe_id"])
    figures = {}

    for probe in probes:

        # get unique channels for this probe
        channels = np.unique(waveforms[waveforms["probe_id"] == probe]["channel_id"])

        # make subplots
        fig = make_subplots(
            rows=(len(channels) + 1) // 2,
            cols=2,
            shared_xaxes=True,
            vertical_spacing=0.1,
            subplot_titles=[f"Channel {channel}" for channel in channels],
        )

        # loop through channels
        for i, channel in enumerate(channels):

            # get units for this channel
            units = waveforms[
                (waveforms["probe_id"] == probe) & (waveforms["channel_id"] == channel)
            ]

            # plot units
            for j, unit in enumerate(units):

                waveform = np.median(unit["waveform"], axis=0)
                q1 = np.percentile(unit["waveform"], 25, axis=0)
                q3 = np.percentile(unit["waveform"], 75, axis=0)
                q1 = waveform + 10
                q3 = waveform - 10
                timestamps = (
                    np.arange(waveform.shape[0]) / (unit["waveform_sample_rate"] / 1000)
                ) - unit["ms_before"]

                color = px.colors.qualitative.Plotly[
                    j % len(px.colors.qualitative.Plotly)
                ]
                fade = 1
                r, g, b = (
                    str(int(fade * int(color[1:3], 16))),
                    str(int(fade * int(color[3:5], 16))),
                    str(int(fade * int(color[5:7], 16))),
                )

                fig.add_trace(
                    go.Scattergl(
                        x=timestamps,
                        y=waveform,
                        mode="lines",
                        name=f"Unit {unit['unit_id']}",
                        legendgroup=f"Unit {unit['unit_id']}",
                        line=dict(color=color),
                        showlegend=True if i == 0 else False,
                    ),  # in the future this will need to be changed if we have sparse extraction
                    row=(i // 2) + 1,
                    col=(i % 2) + 1,
                )

            fig.update_yaxes(
                title_text="Voltage (uV)", row=(i // 2) + 1, col=(i % 2) + 1
            )
            if i // 2 + 1 == len(channels) // 2:
                fig.update_xaxes(
                    title_text="Time (ms)", row=(i // 2) + 1, col=(i % 2) + 1
                )

        fig.update_layout(
            title=f"Probe {probe} waveforms", showlegend=True, width=800, height=800
        )

        figures[int(probe)] = fig

    return figures


@st.cache_data(ttl=3600, max_entries=3)
def plot_isi(spikesorting_dict, selected_units):
    """
    Plot the isi of the selected units
    """

    # open database connection for main thread
    conn = connect.connect()

    # import schema structure
    tables = import_schemas.schema(conn)

    # fetch spiketrains
    spiketrains = (tables["SpikeTrain"] & spikesorting_dict & selected_units).fetch()

    # loop through probes
    probes = np.unique(spiketrains["probe_id"])
    figures = {}

    for probe in probes:

        # get unique channels for this probe
        units = np.unique(spiketrains[spiketrains["probe_id"] == probe]["unit_id"])

        # make subplots
        fig = make_subplots(
            rows=(len(units) + 1) // 2,
            cols=2,
            shared_xaxes=True,
            vertical_spacing=0.1,
            subplot_titles=[f"Unit {unit}" for unit in units],
        )

        # loop through channels
        for i, unit in enumerate(units):

            spiketrain = spiketrains[
                (spiketrains["probe_id"] == probe) & (spiketrains["unit_id"] == unit)
            ]

            # calculate intervals
            isi = np.diff(spiketrain["spiketrain"][0])
            isi = isi[isi > 0]
            isi = isi * 1000
            isi = isi[isi < 10]
            isi = np.append(isi, -1 * isi)

            fig.add_trace(
                go.Histogram(
                    x=isi,
                    name=f"Unit {spiketrain['unit_id']}",
                    legendgroup=f"Unit {spiketrain['unit_id']}",
                    nbinsx=100,
                    histnorm="probability density",
                ),  # in the future this will need to be changed if we have sparse extraction
                row=(i // 2) + 1,
                col=(i % 2) + 1,
            )

            fig.update_yaxes(
                title_text="ISI histogram", row=(i // 2) + 1, col=(i % 2) + 1
            )
            if i // 2 + 1 == len(units) // 2 or len(units) == 1:
                fig.update_xaxes(
                    title_text="Time (ms)",
                    row=(i // 2) + 1,
                    col=(i % 2) + 1,
                    showticklabels=True,
                )
            else:
                fig.update_xaxes(showticklabels=True, row=(i // 2) + 1, col=(i % 2) + 1)
        fig.update_layout(
            title=f"Probe {probe} ISI histograms",
            width=800,
            height=400 * ((len(units) + 1) // 2),
            showlegend=False,
        )
        figures[int(probe)] = fig

    return figures


@st.cache_data(ttl=3600, max_entries=3)
def unit_stats(spikesorting_dict, selected_units):
    """
    Computes some statistics for each unit
    """

    # open database connection for main thread
    conn = connect.connect()

    # import schema structure
    tables = import_schemas.schema(conn)

    # fetch spiketrains
    spiketrains = (tables["SpikeTrain"] & spikesorting_dict & selected_units).fetch()

    # loop through probes
    probes = np.unique(spiketrains["probe_id"])
    figures = {}

    for probe in probes:

        data = []

        T = session_length(spikesorting_dict)

        # get unique channels for this probe
        units = np.unique(spiketrains[spiketrains["probe_id"] == probe]["unit_id"])

        # loop through channels
        for i, unit in enumerate(units):

            spiketrain = spiketrains[
                (spiketrains["probe_id"] == probe) & (spiketrains["unit_id"] == unit)
            ]

            N = len(spiketrain["spiketrain"][0])
            diff = 1000 * np.diff(spiketrain["spiketrain"][0])
            tr = 1.5  # violation threshold in ms
            nv = len(diff[diff < tr])
            C = (nv * T) / (2 * N**2 * tr)  # formula from ultramegasort package

            mean_rate = N / T

            data.append(
                {"Unit": unit, "Mean firing rate (Hz)": mean_rate, "ISI violations": C}
            )

        figures[int(probe)] = pd.DataFrame(data)

    return figures


def plot_probe(probefile):

    probe = pi.io.read_probeinterface(probefile)
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection="3d")
    plot_probe_group(probe, ax=ax)

    return fig



import streamlit as st
import concurrent.futures
from antelope.utils.os_utils import get_config


# function to initialize or get the existing thread pool from session state
def upload_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "upload_futures" not in st.session_state:
        st.session_state.upload_futures = []
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def feature_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "feature_futures" not in st.session_state:
        st.session_state.feature_futures = []
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def behaviour_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "behaviour_futures" not in st.session_state:
        st.session_state.behaviour_futures = []
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def download_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "download_futures" not in st.session_state:
        st.session_state.download_futures = []
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def delete_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "delete_futures" not in st.session_state:
        st.session_state.delete_futures = []
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def restore_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "restore_futures" not in st.session_state:
        st.session_state.restore_futures = []
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def release_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "release_futures" not in st.session_state:
        st.session_state.release_futures = []
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def dlcup_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "upload_dlc_futures" not in st.session_state:
        st.session_state.upload_dlc_futures = []
    if "extract_frames_futures" not in st.session_state:
        st.session_state.extract_frames_futures = []
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def dlc_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "download_dlc_futures" not in st.session_state:
        st.session_state.download_dlc_futures = []
    if "extract_frames_futures" not in st.session_state:
        st.session_state.extract_frames_futures = []
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def phy_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def mask_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "mask_futures" not in st.session_state:
        st.session_state.mask_futures = []
    return st.session_state.process_pool


# function to initialize or get the existing thread pool from session state
def export_thread_pool():
    numworkers = get_config()["multithreading"]["max_workers"]
    if "process_pool" not in st.session_state:
        # create a ThreadPoolExecutor with 1 thread
        st.session_state.process_pool = concurrent.futures.ProcessPoolExecutor(
            max_workers=numworkers
        )
    if "export_futures" not in st.session_state:
        st.session_state.export_futures = []
    return st.session_state.process_pool



"""
This module contains the standard library firing rate analysis.
"""

import numpy as np
from antelope import antelope_analysis
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt


@antelope_analysis
class SpikeCountRate:
    """
    Calculate the spike count rate of a neuron.
    This is the number of spikes divided by the duration of the recording.
    """

    name = "spike_count_rate"
    query = "SpikeTrain"
    returns = {"firing_rate": float}

    def run(key):

        spiketrain = (SpikeTrain & key).fetch1("spiketrain")
        spike_count_rate = spiketrain.shape[0] / spiketrain[-1]

        return spike_count_rate


@antelope_analysis
class FiringRate:
    """
    Smooth the spike train with a kernel.

    Arguments:
    kernel (np.ndarray[2,n]): The kernel to convolve the spike train with. Should consist of values and timestamps.
    sample_rate (float): The sample rate of the spike train.
    """

    name = "firing_rate"
    query = "SpikeTrain"
    returns = {"firing_rate": np.ndarray}
    args = {"kernel": np.ndarray, "sample_rate": float}
    hidden = True

    def run(key, kernel, sample_rate):

        # fetch spiketrain
        spiketrain = (SpikeTrain & key).fetch1("spiketrain")

        # interpolate the kernel to the sample rate
        kernel_func = interp1d(
            kernel[1, :], kernel[0, :], fill_value=0, bounds_error=False
        )
        kernel_interpolated = kernel_func(
            np.arange(kernel[1, 0], kernel[1, -1], 1 / sample_rate)
        )

        # timestamps of output
        timestamps = np.arange(0, spiketrain[-1], 1 / sample_rate)

        # bin the spike train
        spiketrain_binned = np.zeros_like(timestamps, dtype=int)
        bin_edges = timestamps - 1 / (2 * sample_rate)
        spiketrain_binned[np.digitize(spiketrain, bin_edges) - 1] = 1

        # convolve the spike train with the kernel
        firing_rate = np.convolve(spiketrain_binned, kernel_interpolated, mode="same")

        # return firing rate and timestamps
        firing_rate = np.vstack((firing_rate, timestamps))

        return firing_rate


@antelope_analysis
class GaussianSmoothing:
    """
    Smooth the spike train with a Gaussian kernel.

    Arguments:
    sigma (float): The standard deviation of the Gaussian kernel in seconds.
    sample_rate (float): The sample rate of the output firing rate
    truncate (float): The number of standard deviations to include in the kernel.
    """

    name = "gaussian_smoothing"
    query = "SpikeTrain"
    returns = {"smoothed_spiketrain": np.ndarray, "timestamps": np.ndarray}
    args = {"sigma": float, "sample_rate": float, "truncate": float}
    calls = ["firing_rate"]

    def run(key, sigma=0.1, sample_rate=100, truncate=3):

        # compute the kernel
        kernel_timestamps = np.arange(
            -truncate * sigma, truncate * sigma, 1 / sample_rate
        )
        kernel_values = np.exp(-(kernel_timestamps**2) / (2 * sigma**2))
        kernel_values *= sample_rate / np.sum(kernel_values)
        kernel = np.vstack((kernel_values, kernel_timestamps))

        # compute the firing rate
        smoothed_spiketrain = firing_rate(key, kernel, sample_rate)

        return (
            smoothed_spiketrain["firing_rate"][0, :],
            smoothed_spiketrain["firing_rate"][1, :],
        )


@antelope_analysis
class RectangularSmoothing:
    """
    Smooth the spike train with a rectangular kernel.

    Arguments:
    width (float): The width of the kernel in seconds.
    sample_rate (float): The sample rate of the output firing rate.
    """

    name = "rectangular_smoothing"
    query = "SpikeTrain"
    returns = {"smoothed_spiketrain": np.ndarray, "timestamps": np.ndarray}
    args = {"width": float, "sample_rate": float}
    calls = ["firing_rate"]

    def run(key, width=0.1, sample_rate=1000):

        # compute the kernel
        kernel_timestamps = np.arange(-width / 2, width / 2, 1 / sample_rate)
        kernel_values = np.ones_like(kernel_timestamps)
        kernel_values *= sample_rate / np.sum(kernel_values)
        kernel = np.vstack((kernel_values, kernel_timestamps))

        # compute the firing rate
        smoothed_spiketrain = firing_rate(key, kernel, sample_rate)

        return (
            smoothed_spiketrain["firing_rate"][0, :],
            smoothed_spiketrain["firing_rate"][1, :],
        )


@antelope_analysis
class ExponentialSmoothing:
    """
    Smooth the spike train with a exponential kernel.

    Arguments:
    tau (float): The time parameter of the kernel in seconds.
    sample_rate (float): The sample rate of the output firing rate.
    truncate (float): The number of time constants to include in the kernel.
    """

    name = "exponential_smoothing"
    query = "SpikeTrain"
    returns = {"smoothed_spiketrain": np.ndarray, "timestamps": np.ndarray}
    args = {"tau": float, "sample_rate": float, "truncate": float}
    calls = ["firing_rate"]

    def run(key, tau=0.1, sample_rate=1000, truncate=5):

        # compute the kernel
        kernel_timestamps = np.arange(0, tau * truncate, 1 / sample_rate)
        kernel_values = np.exp(-kernel_timestamps / tau)
        kernel_values *= sample_rate / np.sum(kernel_values)
        kernel = np.vstack((kernel_values, kernel_timestamps))

        # compute the firing rate
        smoothed_spiketrain = firing_rate(key, kernel, sample_rate)

        return (
            smoothed_spiketrain["firing_rate"][0, :],
            smoothed_spiketrain["firing_rate"][1, :],
        )


@antelope_analysis
class PlotSmoothed:
    """
    Smooth the spike train with a Gaussian kernel and plot.

    Arguments:
    sigma (float): The standard deviation of the Gaussian kernel in seconds.
    sample_rate (float): The sample rate of the output firing rate
    window_start (float): The start of the window to plot in seconds.
    window_end (float): The end of the window to plot in seconds (default is -1 the end of the recording).
    width (float): The width parameter of the kernel.
    kernel (str): The type of kernel to use ('gaussian', 'rectangular', 'exponential').
    """

    name = "plot_firing_rate"
    query = "SpikeTrain"
    returns = {"plot": plt.Figure}
    args = {
        "sample_rate": int,
        "window_start": float,
        "window_end": float,
        "width": float,
        "kernel": ["gaussian", "rectangular", "exponential"],
    }
    calls = ["gaussian_smoothing", "rectangular_smoothing", "exponential_smoothing"]

    def run(
        key,
        window_start=0,
        window_end=-1,
        width=0.1,
        sample_rate=1000,
        kernel="gaussian",
    ):

        # fetch spiketrain
        spiketrain = (SpikeTrain & key).fetch1("spiketrain")

        # compute the smoothed spiketrain
        if kernel == "gaussian":
            result = gaussian_smoothing(key, width, sample_rate)
            smoothed_spiketrain = result["smoothed_spiketrain"]
            timestamps = result["timestamps"]

        elif kernel == "rectangular":
            result = rectangular_smoothing(key, width, sample_rate)
            smoothed_spiketrain = result["smoothed_spiketrain"]
            timestamps = result["timestamps"]

        elif kernel == "exponential":
            result = exponential_smoothing(key, width, sample_rate)
            smoothed_spiketrain = result["smoothed_spiketrain"]
            timestamps = result["timestamps"]

        # crop to within window
        if window_end == -1:
            window_end = timestamps[-1]
        start_idx = np.argmax(timestamps >= window_start)
        end_idx = np.argmax(timestamps >= window_end)
        smoothed_spiketrain = smoothed_spiketrain[start_idx:end_idx]
        timestamps = timestamps[start_idx:end_idx]
        spiketrain = spiketrain[
            (spiketrain >= window_start) & (spiketrain <= window_end)
        ]

        # plot spiketrain and smoothed spiketrain
        fig, ax = plt.subplots()

        ax.plot(timestamps, smoothed_spiketrain, label="Firing rate")
        ax.plot(spiketrain, np.zeros_like(spiketrain), "|", label="Spikes")
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Firing rate (Hz)")
        ax.set_title("Firing rate of neuron")

        return fig



import streamlit as st
from antelope.utils.streamlit_utils import (
    dropdown_insert_table,
    define_rig_json,
    get_rig_videos,
)
import json
import streamlit_antd_components as sac
from antelope.utils.antelope_utils import validate_behaviour_rig, insert_features


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Behaviour")
        st.subheader("Define Behaviour Rig Layout")

        st.divider()

        sac.buttons(
            [
                sac.ButtonsItem(label="Manually Specify", icon="pen"),
                sac.ButtonsItem(label="Import Json", icon="cloud-upload"),
            ],
            key="behaviour_import",
            gap="md",
            use_container_width=True,
        )

        # manually specify
        if st.session_state["behaviour_import"] == "Manually Specify":

            tablename, insert_dict = dropdown_insert_table(
                tables,
                {"BehaviourRig": tables["BehaviourRig"]},
                username,
                headless=True,
                primary_only=True,
            )

            insert_dict["behaviourrig_name"] = st.text_input("Enter behaviourrig_name")

            st.text("")

            if insert_dict:

                st.divider()

                st.subheader("Define Videos in Rig")

                videos = get_rig_videos()
                videos = list(videos.values())

                st.divider()

                st.subheader("Define Rig Layout")

                rig_json = define_rig_json(videos)

                st.text("")
                if st.button("Insert"):

                    # check user only inserting their own data
                    if insert_dict["experimenter"] == username:

                        value = validate_behaviour_rig(rig_json)

                        if value != True:

                            st.error(value)

                        else:

                            insert_dict["rig_json"] = rig_json

                            # get primary keys
                            primary_key = {
                                key: val
                                for key, val in insert_dict.items()
                                if key in ["experimenter", "behaviourrig_id"]
                            }

                            # make list of features to insert
                            feature_list = insert_features(primary_key, rig_json)

                            # in a single transaction, perform all inserts
                            with tables["Experimenter"].connection.transaction:

                                # insert json into database
                                tables[tablename].insert1(insert_dict)

                                # insert features into database
                                tables["Feature"].insert(feature_list)

                            # show success message
                            st.text("")
                            st.success("Data uploaded to database!")

                    # otherwise print error
                    else:

                        st.text("")
                        st.error("You can only insert your own data!")

        # import predefined json
        elif st.session_state["behaviour_import"] == "Import Json":

            tablename, insert_dict = dropdown_insert_table(
                tables,
                {"BehaviourRig": tables["BehaviourRig"]},
                username,
                headless=True,
            )

            st.text("")

            if insert_dict:

                if st.button("Insert"):

                    # check user only inserting their own data
                    if insert_dict["experimenter"] == username:

                        try:

                            # first, validate the behaviour rig file
                            rigfile = insert_dict["rig_json"]

                            with open(rigfile, "r") as f:
                                read_rig = json.load(f)

                            # validate the rig file
                            assert validate_behaviour_rig(read_rig)

                        except AssertionError:

                            st.error("Rig file not valid!")
                            st.warning(
                                "Please read the documentation and ensure your file is in the correct format"
                            )

                        except Exception:
                            st.error("Error reading file!")

                        else:

                            insert_dict["rig_json"] = read_rig

                            # get primary keys
                            primary_key = {
                                key: val
                                for key, val in insert_dict.items()
                                if key in ["experimenter", "behaviourrig_id"]
                            }

                            # make list of features to insert
                            feature_list = insert_features(primary_key, read_rig)

                            # in a single transaction, perform all inserts
                            with tables["Experimenter"].connection.transaction:

                                # insert json into database
                                tables[tablename].insert1(insert_dict)

                                # insert features into database
                                tables["Feature"].insert(feature_list)

                            # show success message
                            st.text("")
                            st.success("Data uploaded to database!")

                    # otherwise print error
                    else:

                        st.text("")
                        st.error("You can only insert your own data!")



import os
from pathlib import Path
from antelope.utils.os_utils import cp_st_config
import runpy
import sys
import warnings


def run():

    try:

        import streamlit

    except ImportError:

        print(
            """
Antelope GUI is not installed. Please install using:

pip install antelope[gui]

Or run the command line version using:

antelope-python
        """
        )

    else:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # copy streamlit config to home if it doesn't exist
            cp_st_config()

            app = Path(os.path.abspath(__file__)).parent / "app.py"

            sys.argv = ["streamlit", "run", str(app)]
            runpy.run_module("streamlit", run_name="__main__")



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_query_table
from antelope.utils.external_utils import schedule_inference_dlc, inference_dlc_progress


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Behaviour")
        st.subheader("Extract kinematics from videos")

        st.divider()

        query = (tables["World"] & {"world_deleted": "False"}) & (
            tables["DLCModel"] & {"dlcmodel_deleted": "False"}
        ) - tables["Kinematics"]

        # Select the feature to add
        tablename, key = dropdown_query_table(
            tables, {"World": query}, username, headless=True
        )

        if key is None:

            st.warning(
                "You don't have any unprocessed videos or don't yet have a trained deeplabcut model."
            )

        else:

            st.divider()
            st.subheader("Check your data")

            query = (
                (
                    tables["World"]
                    & {"world_deleted": "False"}
                    & (tables["DLCModel"] & {"dlcmodel_deleted": "False"})
                    - tables["Kinematics"]
                    & key
                )
                .aggr(
                    (
                        tables["World"] * tables["DLCModel"] - tables["Kinematics"]
                        & {
                            "world_deleted": "False",
                            "dlcmodel_deleted": "False",
                            "kinematics_deleted": "False",
                        }
                    ).proj(),
                    number_models="count(*)",
                )
                .proj("number_models")
            )

            if len(query) == 0:
                st.warning("There is nothing to be processed.")
                return

            df = query.fetch(format="frame")
            st.dataframe(df, use_container_width=True)
            total = df["number_models"].sum()
            st.write(f"Total number of inference jobs to run: {total}")

            st.divider()

            # schedule computation section
            st.subheader("Schedule computation")

            # ask for user password for scheduling the job
            password = st.text_input(
                "Please enter your cluster password", type="password"
            )

            st.text("")
            if st.button("Schedule deeplabcut inference"):

                if key["experimenter"] != username:

                    st.error("You can only train models on your own data")

                else:

                    try:

                        num_videos = total

                        # send job
                        schedule_inference_dlc(key, num_videos, password)

                        # success message
                        st.success("Job sent to cluster!")
                        st.info("You will receive an email once your job is completed")

                    except Exception as e:

                        st.error("Error submitting job to cluster")
                        print(e)

            # if there are any downlaods this session
            if "inference_dlc_jobs" in st.session_state:

                # button which shows spikesort statuses
                if st.button("Check model inference progress"):

                    inference_dlc_progress()



import streamlit as st
from antelope.utils.streamlit_utils import (
    dropdown_query_table,
    server_directory_browser,
)
from antelope.utils.datajoint_utils import upload
from antelope.utils.multithreading_utils import feature_thread_pool
import pandas as pd


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Behaviour")
        st.subheader("Add features for the behaviour rig")

        st.divider()

        # Select the feature to add
        tablename, key = dropdown_query_table(
            tables, {"Feature": tables["Feature"]}, username, headless=True
        )

        # if no features in database raise warning
        if tablename == None:
            st.warning(
                "You need to add a behaviour rig to the database before adding features"
            )

        else:

            # Query if there's already a feature attached
            query = tables["Feature"] & key & {"feature_data": None}

            if len(query) == 0:
                st.text("")
                st.warning(
                    "This feature already has data attached. Uploading data again will overwrite the existing data, so make sure you want to update the feature."
                )

            # Get user to select the feature data
            if "feature_id" in key and key["experimenter"] == username:

                st.divider()

                # get feature data
                feature_data = server_directory_browser("Select feature data", None)

                # add to key
                key["feature_data"] = feature_data

                if feature_data:

                    st.divider()

                    # make insert button
                    if st.button("Insert"):

                        # retrieve thread pool
                        up_thread_pool = feature_thread_pool()

                        # submit job to thread pool
                        future = up_thread_pool.submit(
                            upload, *(tablename, key, "update")
                        )

                        # append future to session state
                        st.session_state.feature_futures.append(
                            (future, tablename, key)
                        )
                        st.text("")
                        st.success("Upload in progress!")

            elif key["experimenter"] != username:

                st.error("You can only insert your own data!")

            # print warning
            st.text("")
            st.info(
                "Note that uploading large features can take a while. This will occur in a separate thread so you can still use Antelope while the upload is occurring, and can use the button below to check your upload status."
            )

            st.text("")

            if "feature_futures" in st.session_state:

                if st.button("Check insert progress"):

                    # if there are any downloads this session
                    if "feature_futures" in st.session_state:
                        st.write("Upload statuses:")

                        # initialise data
                        display_futures = []

                        # compute job statuses
                        for (
                            future,
                            tablename,
                            insert_dict,
                        ) in st.session_state.feature_futures:

                            # compute statuses
                            if future.done():
                                if future.exception():
                                    status = "upload error"
                                else:
                                    status = "upload success"
                            else:
                                status = "upload in progress"

                            # primary keys for display
                            keys = {
                                key: val
                                for key, val in insert_dict.items()
                                if key in tables[tablename].primary_key
                            }
                            display = "-".join([str(i) for i in keys.values()])

                            display_futures.append((tablename, display, status))

                        # make dataframe to display
                        df = pd.DataFrame(
                            display_futures, columns=["Table", "Primary Key", "Status"]
                        )

                        # show dataframe
                        st.dataframe(df, hide_index=True)

                    # if there are no downloads in this session
                    else:
                        st.write("No uploads underway.")



import streamlit as st
from antelope.utils.streamlit_utils import (
    dropdown_insert_table,
    server_directory_browser,
)
from antelope.utils.visualisation_utils import plot_probe
import probeinterface as pi
import json
import streamlit_antd_components as sac
from pathlib import Path
import os


def validate_probe(probefile):
    read_probe = pi.io.read_probeinterface(probefile)
    assert all([probe["ndim"] == 3 for probe in read_probe.to_dict()["probes"]])


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Electrophysiology")
        st.subheader("Insert Probe Geometry File")

        st.divider()

        # get user to interactively insert table attributes
        tablename, insert_dict = dropdown_insert_table(
            tables,
            {"ProbeGeometry": tables["ProbeGeometry"]},
            username,
            headless=True,
            primary_only=True,
        )

        # get json file
        st.text("")
        option = sac.buttons(
            [
                sac.ButtonsItem(label="Local File", icon="filetype-json"),
                sac.ButtonsItem(label="Standard Library", icon="archive-fill"),
            ],
            use_container_width=True,
        )

        if option == "Local File":
            insert_dict["probe"] = server_directory_browser(
                "Select probe geometry file", "json"
            )
        elif option == "Standard Library":
            resources = (
                Path(os.path.abspath(__file__)).parent.parent.parent / "resources"
            )
            relative_resources = Path(os.path.relpath(resources, Path.cwd()))
            options = {f.stem: f for f in resources.glob("probes/*.json")}
            insert_dict["probe"] = options[
                st.selectbox("Select probe geometry file", list(options.keys()))
            ]
        insert_dict["probegeometry_name"] = st.text_input("Enter probe geometry name")

        st.text("")
        col4, col5 = st.columns([0.5, 0.5])

        with col4:
            if st.button("Plot Probe", use_container_width=True):
                try:
                    probefile = insert_dict["probe"]
                    validate_probe(probefile)
                except AssertionError:
                    with col2:
                        st.error("Probe geometry must be 3D!")
                except Exception:
                    with col2:
                        st.error("Probe geometry file not valid!")
                else:
                    st.session_state["plot_probe"] = plot_probe(probefile)

        with col5:
            if st.button("Insert", use_container_width=True):

                # check user only inserting their own data
                if insert_dict["experimenter"] == username:

                    try:

                        # first, validate probe geometry file
                        probefile = insert_dict["probe"]

                        validate_probe(probefile)

                    except AssertionError:

                        with col2:
                            st.error("Probe geometry must be 3D!")

                    except Exception:

                        with col2:
                            st.error("Probe geometry file not valid!")

                    else:
                        # insert into database
                        with open(probefile, "r") as f:
                            insert_dict["probe"] = json.load(f)

                        tables[tablename].insert1(insert_dict)

                        # show success message
                        with col2:
                            st.text("")
                            st.success("Data uploaded to database!")

                # otherwise print error
                else:

                    with col2:
                        st.text("")
                        st.error("You can only insert your own data!")

        if hasattr(st.session_state, "plot_probe"):
            st.pyplot(st.session_state["plot_probe"])



"""
Script contains functions to enforce user permissions.
These will overwrite the table methods via monkey patching.
"""

from antelope.utils.datajoint_utils import delete_restriction
from datajoint.table import Table


class PermissionDeniedError(Exception):
    """Exception raised when a user lacks permission to perform an action."""

    def __init__(self, message="You do not have permission to delete this data."):
        self.message = message
        super().__init__(self.message)


def modify_insert1(method):
    def wrapper(self, row, **kwargs):
        if hasattr(self, "usertable"):
            for key in self.heading.primary_key:
                if (
                    self.heading.attributes[key].comment[-16:] == "(auto_increment)"
                    and key not in row.keys()
                ):
                    row[key] = max((self.proj() & row).fetch(key), default=0) + 1
            return method(self, row, **kwargs)
        else:
            return method(self, row, **kwargs)

    return wrapper


Table.insert1 = modify_insert1(Table.insert1)


def modify_insert(method):
    def wrapper(
        self,
        rows,
        replace=False,
        skip_duplicates=False,
        ignore_extra_fields=False,
        allow_direct_insert=None,
    ):
        """
        Performs insert after checking user permissions
        """
        if hasattr(self, "usertable"):
            username = self.connection.get_user().split("@")[0]
            admin = (self.tables["Experimenter"] & {"experimenter": username}).fetch1(
                "admin"
            )

            if admin == "False":
                if list((self.tables["Experimenter"] & rows).fetch("experimenter")) != [
                    username
                ]:
                    raise PermissionDeniedError(
                        "You do not have permission to insert this data."
                    )
                else:
                    return method(
                        self,
                        rows,
                        replace,
                        skip_duplicates,
                        ignore_extra_fields,
                        allow_direct_insert,
                    )
            elif admin == "True":
                return method(
                    self,
                    rows,
                    replace,
                    skip_duplicates,
                    ignore_extra_fields,
                    allow_direct_insert,
                )
        else:
            return method(
                self,
                rows,
                replace,
                skip_duplicates,
                ignore_extra_fields,
                allow_direct_insert,
            )

    return wrapper


Table.insert = modify_insert(Table.insert)


def modify_update1(method):
    def wrapper(self, row):
        """
        Performs update1 after checking user permissions
        """
        if hasattr(self, "usertable"):
            username = self.connection.get_user().split("@")[0]
            admin = (self.tables["Experimenter"] & {"experimenter": username}).fetch1(
                "admin"
            )

            if admin == "False":
                if "experimenter" not in row.keys() or row["experimenter"] != username:
                    raise PermissionDeniedError(
                        "You do not have permission to update this data."
                    )
                else:
                    return method(self, row)
            elif admin == "True":
                return method(self, row)
        else:
            return method(self, row)

    return wrapper


Table.update1 = modify_update1(Table.update1)


def modify_delete(quick_delete, delete):
    def wrapper(self, safemode=True):
        """
        Wrapper around datajoint's delete that checks user is an admin first
        """
        if hasattr(self, "usertable"):
            # first, check permissions
            username = self.connection.get_user().split("@")[0]
            admin = (self.tables["Experimenter"] & {"experimenter": username}).fetch1(
                "admin"
            )

            if admin == "False":
                raise PermissionDeniedError(
                    "You do not have permission to perform a true delete. Please use temp_delete instead."
                )
            elif admin == "True":
                if safemode:
                    return delete(self)
                else:
                    return quick_delete(self)
        else:
            if safemode:
                return delete(self)
            else:
                return quick_delete(self)

    return wrapper


Table.delete = modify_delete(Table.delete_quick, Table.delete)


def safe_delete_python(query):
    """
    Safely deletes data from the database (just changes the 'delete' attribute to True)
    These deletes cascade
    Inputs: query: the query to delete
    """

    # first, check permissions
    username = query.connection.get_user().split("@")[0]
    admin = (query.tables["Experimenter"] & {"experimenter": username}).fetch1("admin")

    # disallow if user not admin and modifying other user's data
    if (
        admin == "False"
        and f"""(`experimenter`="{username}")""" not in query.make_sql()
    ):
        raise PermissionDeniedError()

    else:
        # all updates in transaction
        update_dict = {}
        full_names = {val.full_table_name: key for key, val in query.tables.items()}
        with query.connection.transaction:

            # loop through all tables to be modified
            for tablename in query.descendants():

                # retrieve datajoint table from full name
                table = query.tables[full_names[tablename]]

                # main query
                full_query = table & (query.proj() & delete_restriction(query, "False"))

                # fetch primary keys of rows to be deleted
                data = full_query.proj().fetch(as_dict=True)

                update_dict[tablename] = data

            # now loop again to perform modifications
            for tablename, data in update_dict.items():

                # retrieve datajoint table from full name
                table = query.tables[full_names[tablename]]

                for i in data:
                    table.update1({**i, **delete_restriction(table, "True")})


Table.tmp_delete = safe_delete_python



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_insert_table
from antelope.utils.datajoint_utils import upload, check_session
from antelope.utils.multithreading_utils import upload_thread_pool
import pandas as pd


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Session")
        st.subheader("Upload a raw electrophysiology recording")

        st.divider()

        # get user to interactively insert table attributes
        tablename, insert_dict = dropdown_insert_table(
            tables, {"Recording": tables["Recording"]}, username, headless=True
        )

        # if upstream tables not populated, print error
        if tablename is None:

            st.text("")
            st.error(
                """You can't upload a recording yet as you haven't got entries in the necessary upstream tables."""
            )
            st.warning(
                "Please make sure you have a session for which you can upload a recording."
            )
            st.stop()

        st.text("")

        # add insert button
        if st.button("Insert"):

            # check user only inserting their own data
            if insert_dict["experimenter"] == username:

                # if directory has correct files and there's not already a recording, upload in separate thread
                status = (
                    len(
                        tables["Recording"]
                        & {
                            key: val
                            for key, val in insert_dict.items()
                            if key
                            in [
                                "experimenter",
                                "experiment_id",
                                "animal_id",
                                "session_id",
                            ]
                        }
                    )
                    == 0
                )
                if check_session(tables, insert_dict) and status:

                    # retrieve thread pool
                    up_thread_pool = upload_thread_pool()

                    # submit job to thread pool
                    future = up_thread_pool.submit(upload, *(tablename, insert_dict))

                    # append future to session state
                    st.session_state.upload_futures.append(
                        (future, tablename, insert_dict)
                    )
                    st.text("")
                    st.success("Upload in progress!")

                elif not status:
                    st.text("")
                    st.error("A recording already exists for this session.")

                # otherwise print error
                else:
                    equip_type = (tables["Experiment"] & insert_dict).fetch1(
                        "ephys_acquisition"
                    )
                    st.text("")
                    st.error("Recording directory does not contain correct files!")
                    st.text("")
                    st.warning(
                        f"Please read the {equip_type} documentation to see what files are required."
                    )

            # otherwise print error
            else:

                st.text("")
                st.error("You can only insert your own data!")

        st.text("")

        # notice
        st.info(
            "Note that uploading large session recordings can take a while. This will occur in a separate thread so you can still use Antelope while the upload is occurring, and can use the button below to check your upload status."
        )

        st.text("")

        # add a button which shows upload statuses
        # uses all uploads in current session stored in session state
        # only shows if there are any downloads in this session

        if "upload_futures" in st.session_state:

            if st.button("Check insert progress"):

                # if there are any downloads this session
                if "upload_futures" in st.session_state:
                    st.write("Upload statuses:")

                    # initialise data
                    display_futures = []

                    # compute job statuses
                    for (
                        future,
                        tablename,
                        insert_dict,
                    ) in st.session_state.upload_futures:

                        # compute statuses
                        if future.done():
                            if future.exception():
                                status = "upload error"
                            else:
                                status = "upload success"
                        else:
                            status = "upload in progress"

                        # primary keys for display
                        keys = {
                            key: val
                            for key, val in insert_dict.items()
                            if key in tables[tablename].primary_key
                        }
                        display = "-".join([str(i) for i in keys.values()])

                        display_futures.append((tablename, display, status))

                    # make dataframe to display
                    df = pd.DataFrame(
                        display_futures, columns=["Table", "Primary Key", "Status"]
                    )

                    # show dataframe
                    st.dataframe(df, hide_index=True)

                # if there are no downloads in this session
                else:
                    st.write("No uploads underway.")



import streamlit as st
from datajoint.user_tables import TableMeta
from antelope.utils.datajoint_utils import (
    query_without_external,
    download_data,
    safe_delete,
    form_query,
    show_deletes,
    query_to_str,
    delete_status,
    true_delete,
)
from antelope.utils.streamlit_utils import (
    dropdown_query_table,
    server_directory_browser,
    children_buttons,
    go_back,
)
from antelope.utils.multithreading_utils import delete_thread_pool, download_thread_pool
import pandas as pd
import streamlit_antd_components as sac


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Search Database")
        st.subheader("Search across all schemas")

        st.divider()

        def top_chg():
            if hasattr(st.session_state, "search_fct"):
                del st.session_state.search_fct

        sac.buttons(
            [
                sac.ButtonsItem(label="Sequential Filter", icon="sort-down"),
                sac.ButtonsItem(label="Manual Query", icon="pen"),
            ],
            key="search_mode",
            gap="md",
            use_container_width=True,
            on_change=top_chg,
        )

        st.text("")

        if st.session_state["search_mode"] == "Sequential Filter":

            def reset_tablename():
                del st.session_state.tablename

            # get schema
            if hasattr(st.session_state, "go_back"):
                if st.session_state.go_back:
                    st.session_state.schema = st.session_state.tmp_schema
                    st.session_state.go_back = False
            sac.buttons(
                [
                    sac.ButtonsItem(label="Metadata", icon="table"),
                    sac.ButtonsItem(label="Electrophysiology", icon="lightning"),
                    sac.ButtonsItem(label="Behaviour", icon="joystick"),
                ],
                key="schema",
                use_container_width=True,
                on_change=reset_tablename,
            )

            # make table mapping
            table_mapping = {
                "Metadata": ["Experimenter", "Experiment", "Animal", "Session"],
                "Electrophysiology": [
                    "ProbeGeometry",
                    "ProbeInsertion",
                    "SortingParams",
                    "Recording",
                    "SpikeSorting",
                    "Probe",
                    "Channel",
                    "LFP",
                    "Unit",
                    "SpikeTrain",
                    "Waveform",
                ],
                "Behaviour": [
                    "BehaviourRig",
                    "LabelledFrames",
                    "DLCModel",
                    "MaskFunction",
                    "Feature",
                    "World",
                    "Video",
                    "Self",
                    "Object",
                    "DigitalEvents",
                    "AnalogEvents",
                    "IntervalEvents",
                    "Kinematics",
                    "Mask",
                ],
            }

            # get input restriction
            subtables = {
                key: val
                for key, val in tables.items()
                if key in table_mapping[st.session_state["schema"]]
            }
            tablename, restriction = dropdown_query_table(
                tables, subtables, username, search_page=True
            )

            if tablename == None:

                st.text("")
                st.error("No tables to search in this schema")
                st.stop()

            else:
                # form datajoint querystring
                querystring = f"{tablename} & {str(restriction)}"

                # form query
                query = form_query(querystring, tables)

        elif st.session_state["search_mode"] == "Manual Query":

            # get manual query
            querystring = st.text_input("Enter your manual query")

            # form query
            if querystring:
                query = form_query(querystring, tables)

        st.text("")

        if st.session_state.search_mode == "Sequential Filter":
            index = ["Data", "Navigation"].index(
                st.session_state.get("search_fct", "Data")
            )
            sac.buttons(
                [
                    sac.ButtonsItem(label="Data"),
                    sac.ButtonsItem(label="Navigation"),
                ],
                key="search_fct",
                use_container_width=True,
                index=index,
            )
        elif st.session_state.search_mode == "Manual Query":
            sac.buttons(
                [
                    sac.ButtonsItem(label="Data"),
                ],
                key="search_fct",
                use_container_width=True,
            )

        st.text("")

        if querystring:

            if st.session_state["search_fct"] == "Data":

                # query results to display
                df, number = query_without_external(
                    query, st.session_state["search_mode"]
                )
                del_col = [False] * len(df)
                df.insert(0, "delete", del_col)

                # show table
                if number == 0:
                    st.warning("No entries to show!")
                else:
                    disabled = df.columns.to_list()
                    disabled.remove("delete")
                    config = {"delete": st.column_config.CheckboxColumn("delete")}
                    delete_df = st.data_editor(
                        df,
                        column_config=config,
                        hide_index=True,
                        use_container_width=True,
                        disabled=disabled,
                    )
                    st.text("Number of items: {}".format(number))

            elif st.session_state["search_fct"] == "Navigation":

                children_buttons(query, tables)
                # query results to display
                delete_df, number = query_without_external(
                    query, st.session_state["search_mode"]
                )
                del_col = [False] * len(delete_df)
                delete_df.insert(0, "delete", del_col)
                if st.button("Back"):
                    go_back()

        st.divider()

        # download and delete section
        sac.buttons(
            [
                sac.ButtonsItem(label="Download", icon="cloud-download"),
                sac.ButtonsItem(label="Delete", icon="trash"),
            ],
            key="mode2",
            gap="md",
            use_container_width=True,
        )

        # download section
        if st.session_state["mode2"] == "Download":

            st.header("Download results")

            # leave blank if no query yet
            if not querystring:

                st.warning("No query entered yet")

                return

            # get user to input download path
            download_path = server_directory_browser("Select download location:")
            filename = st.text_input("Enter filename") + ".npy"
            download_path = download_path / filename

            st.text("")

            # download data
            if st.button("Download"):

                # retrieve thread pool
                down_thread_pool = download_thread_pool()

                # make name for query
                query = (
                    query & True
                )  # protection seems to help download more complex queries
                query_name = query_to_str(query)

                # if recording or video, download in separate thread
                if query.table_name in ["recording", "_video"]:

                    # submit job to thread pool
                    future = down_thread_pool.submit(
                        download_data,
                        *(querystring, download_path, st.session_state["search_mode"]),
                    )

                    # append future to session state
                    st.session_state.download_futures.append((future, query_name))

                    st.success("Download in progress.")

                else:

                    # download in main thread
                    download_data(
                        querystring, download_path, st.session_state["search_mode"]
                    )

                    st.success("Download complete.")

            st.text("")

            # notice
            st.info(
                "Note that downloading large session recordings can take a while. This will occur in a separate thread so you can still use Antelope while the download is occurring, and can use the button below to check your download status."
            )
            st.warning(
                "If you try to download the same data to the same location again before the first download is finished you will potentially corrupt your downloaded data."
            )

            st.text("")

            # add a button which shows download statuses
            # uses all downloads in current session stored in session state
            if "download_futures" in st.session_state:

                if st.button("Check download progress"):

                    # if there are any downlaods this session
                    if "download_futures" in st.session_state:

                        st.write("Download statuses:")

                        # initialise data
                        display_futures = []

                        # display job progress
                        for future, query_name in st.session_state.download_futures:

                            # compute statuses
                            if future.done():
                                if future.exception():
                                    status = "download error"
                                else:
                                    status = "download success"
                            else:
                                status = "download in progress"

                            display_futures.append((query_name, status))

                        # make dataframe to display
                        df = pd.DataFrame(display_futures, columns=["Query", "Status"])

                        # show dataframe
                        st.dataframe(df, hide_index=True)

                    # if there are no downloads in this session
                    else:
                        st.write("No downloads underway.")

        # delete section - only for sequential filter
        if st.session_state["mode2"] == "Delete":

            st.header("Delete data")

            # leave blank if no query yet
            if not querystring:

                st.warning("No query entered yet")

            else:

                query = form_query(querystring, tables)

                # print a warning if manual query and add additional restriction to query
                if st.session_state["search_mode"] == "Manual Query":
                    st.warning(
                        "Note for manual queries, we automatically add the additional restriction that you can only delete your own data. This will be reflected when you check deletes, so the number of entries here will potentially differ from your full search."
                    )

                    query = query & {"experimenter": username}

                # for sequential filters, check experimenter is the user
                if st.session_state["search_mode"] == "Sequential Filter" and (
                    "experimenter" not in restriction.keys()
                    or not restriction["experimenter"] == username
                ):

                    st.error("""You don't have permission to delete this data""")

                # don't want to delete from experimenter table
                elif isinstance(query, tables["Experimenter"].__class__):

                    st.error("You can't delete from the experimenter table")

                # some manual queries, such as joins, are too complex to delete
                # simple queries are our user defined tables or TableMeta type
                elif not isinstance(
                    query, tuple([t.__class__ for t in tables.values()] + [TableMeta])
                ):

                    st.error("This manual query is too complex to delete from")
                    st.warning(
                        "Typically, you can only delete restricted entries from a single table"
                    )

                else:

                    # check delete mode
                    full_names = {
                        val.full_table_name: key for key, val in tables.items()
                    }
                    tablename = full_names[query.full_table_name]
                    delete_mode = delete_status(tablename, tables, query)

                    if delete_mode == "none":

                        st.error("This table cannot be deleted from")

                    else:

                        st.text("")

                        st.info(
                            "Note that deletes cascade to avoid orphaned database entries. Please use the button below to check which tables will get deleted from."
                        )

                        st.text("")

                        cols = query.proj().heading.names
                        to_delete = delete_df[delete_df["delete"] == True][
                            cols
                        ].to_dict(orient="records")

                        if delete_mode == "custom-behave":

                            # change query to world and query
                            querystring = "World & " + querystring.split("&")[1]
                            query = form_query(querystring, tables)
                            delete_mode = "perm"

                        if delete_mode == "perm":

                            st.error(
                                "This delete will be permanent and cannot be undone."
                            )
                            st.text("")

                        delete_query = query & to_delete
                        delete_querystring = querystring + f" & {to_delete}"
                        print(delete_querystring)

                        if st.button("Check deletes"):

                            # show cascaded deletes
                            descendant_dict = show_deletes(tables, delete_query)
                            st.write("Entries that will get deleted:")
                            st.dataframe(
                                pd.DataFrame(data=descendant_dict).set_index("Table")
                            )

                        status = (
                            st.text_input(
                                label="Please enter your password to confirm computation release:",
                                type="password",
                            )
                            == st.session_state.password
                        )

                        st.text("")

                        # delete button
                        if st.button("Delete"):

                            # delete only if password correct
                            if status:

                                if delete_mode == "temp":

                                    # get cascaded deletes
                                    descendant_dict = show_deletes(tables, delete_query)

                                    # count total number of entries to delete
                                    total_deletes = sum(
                                        descendant_dict["Number entries to be deleted"]
                                    )

                                    # use main thread if less than 1000 entries
                                    if total_deletes < 1000:

                                        # delete data
                                        safe_delete(delete_querystring)

                                        st.text("")

                                        st.success("Delete complete.")

                                    # otherwise use thread pool
                                    else:

                                        # retrieve thread pool
                                        del_thread_pool = delete_thread_pool()

                                        # submit job to thread pool
                                        future_delete = del_thread_pool.submit(
                                            safe_delete, delete_querystring
                                        )

                                        # form query name
                                        query_name = query_to_str(delete_query)

                                        # append future to session state
                                        st.session_state.delete_futures.append(
                                            (future_delete, query_name)
                                        )

                                        st.success("Deletion in progress.")

                                elif delete_mode == "perm":

                                    # delete data
                                    true_delete(delete_query)

                            # incorrect password
                            else:
                                st.error("Password incorrect.")

                    st.text("")

                    # notice
                    st.info(
                        "Note that temporary deletes can take a while if there are a very large number of rows to update. This will occur in a separate thread so you can still use Antelope while the delete is occurring, and can use the button below to check your delete status."
                    )

                    st.text("")

                    # add a button which shows download statuses
                    # uses all downlaods in current session stored in session state
                    if "delete_futures" in st.session_state:

                        if st.button("Check delete progress"):

                            # if there are any downlaods this session
                            if (
                                "delete_futures" in st.session_state
                                and len(st.session_state.delete_futures) > 0
                            ):

                                st.write("Delete progress:")

                                # initialise data
                                display_futures = []

                                # display job progress
                                for (
                                    future,
                                    queryname,
                                ) in st.session_state.delete_futures:

                                    # compute statuses
                                    if future.done():
                                        if future.exception():
                                            status = "delete error"
                                        else:
                                            status = "delete success"
                                    else:
                                        status = "delete in progress"

                                    display_futures.append((queryname, status))

                                # make dataframe to display
                                df = pd.DataFrame(
                                    display_futures, columns=["Query", "Status"]
                                )

                                # show dataframe
                                st.dataframe(df, hide_index=True)

                            # if there are no downloads in this session
                            else:
                                st.write("No deletes underway.")



import streamlit as st
from antelope.utils.streamlit_utils import (
    dropdown_query_table,
    display_sorting_parameters,
    lfp_params_input,
    waveform_params_input,
    preprocessing_params_input,
    input_sorter_params,
    agreement_params_input,
)


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Electrophysiology")
        st.subheader("Insert Spikesorting Parameters")

        st.divider()
        st.subheader("Select the animal you want to add parameters to")

        # get input restriction
        _, primary_key = dropdown_query_table(
            tables, {"Animal": tables["Animal"]}, username, headless=True
        )

        # error if prerequisites not satisfied
        if primary_key == None:
            st.error(
                """You can't insert any sorting parameters yet as you haven't got entries in the necessary upstream tables."""
            )
            st.warning(
                "Please make sure you have at least one animal to add sorting parameters to."
            )

        # otherwise retrieve all the data
        elif "animal_id" in primary_key.keys():

            # calculate sortingparams id
            query = tables["SortingParams"] & primary_key
            next_key = max(query.fetch("sortingparams_id"), default=0) + 1
            primary_key["sortingparams_id"] = next_key

            st.divider()

            st.subheader("Insert parameters")

            # initialise paramdict
            paramdict = {}

            """
            LFP parameters
            """
            st.divider()

            # display message
            st.markdown("##### Select LFP extraction parameters:")
            paramdict["lfp"] = lfp_params_input()

            """
            waveform parameters
            """
            st.divider()

            # display message
            st.markdown("##### Select waveform extraction parameters:")
            paramdict["waveform"] = waveform_params_input()

            """
            Preprocessing parameters
            """
            st.divider()

            # display message
            st.markdown("##### Select preprocessing parameters:")

            # get user input preprocessing params
            paramdict["preprocessing"] = preprocessing_params_input()

            """
            Spikesorting parameters
            """
            st.divider()

            # display message
            st.markdown("##### Select spikesorting parameters:")
            st.write(
                "You can select one or more spikesorters which will run in parallel."
            )
            st.write(
                "Note some of the parameters for a given sorter may be irrelevant,\nparticularly those related to the allocation of computational resources."
            )

            # get user input sorter params
            paramdict["spikesorters"] = input_sorter_params()

            # only add agreement matching if more than one sorter
            if len(paramdict["spikesorters"]) > 1:
                """
                Agreement matching parameters
                """
                st.divider()

                # display message
                st.markdown("##### Select agreement matching parameters:")

                # get user input agreement matching params
                paramdict["matching"] = agreement_params_input(
                    len(paramdict["spikesorters"])
                )

            """
            Description
            """
            st.divider()

            st.subheader("Add a name")

            name = st.text_input(
                "Add an identifiable name for this set of sorting parameters"
            )

            st.subheader("Add a description")

            description = st.text_area(
                "Add a description of this set of sorting parameters"
            )

            """
            Upload section
            """
            st.divider()

            st.subheader("Add to database")

            # display confirmation showing parameters to be inserted
            st.text("Adding the following parameters to be run for this experiment:")
            display_sorting_parameters(paramdict)

            # upload on button click
            if st.button("Add to database"):

                # can't have parameters with no spikesorters
                if len(paramdict["spikesorters"]) == 0:

                    st.error("You must add at least one spikesorter!")

                # if user doesn't have permsission
                elif primary_key["experimenter"] != username:

                    st.error("You can only insert your own data!")

                # otherwise data can be inserted
                else:
                    tables["SortingParams"].insert1(
                        {
                            **primary_key,
                            "sortingparams_name": name,
                            "params": paramdict,
                            "sortingparams_notes": description,
                        }
                    )

                    # print success message
                    st.success("Parameters uploaded to database!")



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_insert_table


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Metadata")
        st.subheader("Insert data into the metadata schema")

        st.divider()

        # get dictionary of all session tables names and objects
        if (
            len(
                tables["Experiment"]
                & {"experimenter": username, "experiment_deleted": "False"}
            )
            > 0
        ):
            mantables = {
                key: tables[key] for key in ["Experiment", "Animal", "Session"]
            }
        else:
            mantables = {key: tables[key] for key in ["Experiment"]}

        # get user to interactively insert table attributes
        tablename, insert_dict = dropdown_insert_table(tables, mantables, username)

        st.text("")

        # add insert button
        if st.button("Insert"):

            # check user only inserting their own data
            if insert_dict["experimenter"] == username:

                # insert into database
                tables[tablename].insert1(insert_dict)

                # show success message
                st.text("")
                st.success("Data uploaded to database!")

            # otherwise print error
            else:

                st.text("")
                st.error("You can only insert your own data!")



import probeinterface as pi
from probeinterface import Probe, ProbeGroup
from collections import namedtuple

# neuropixels 1.0
specs = namedtuple("specs", ["num_contacts", "xpitch", "ypitch"])
np1 = {
    "short": specs(960, 32, 20),
    "medium": specs(2496, 103, 20),
    "long": specs(4416, 103, 20),
}
for k, v in np1.items():
    probegroup = ProbeGroup()
    probe = pi.generate_multi_columns_probe(
        num_columns=2,
        num_contact_per_column=v.num_contacts // 2,
        xpitch=v.xpitch,
        ypitch=v.ypitch,
        contact_shapes="square",
        contact_shape_params={"width": 12},
    )
    probegroup.add_probe(probe.to_3d(axes="xz"))
    pi.io.write_probeinterface(
        f"../resources/probes/neuropixels1.0_{k}.json", probegroup
    )

# tetrodes
n_tetrodes = [1, 2, 4, 8, 16]
for n in n_tetrodes:
    probegroup = ProbeGroup()
    for i in range(n):
        probe = pi.generate_tetrode()
        probe.move([i * 50, 0])
        probegroup.add_probe(probe.to_3d(axes="xz"))
    pi.io.write_probeinterface(f"../resources/probes/tetrode_{n}.json", probegroup)



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_query_table
from antelope.utils.datajoint_utils import safe_delete, query_to_str, show_deletes
from antelope.utils.external_utils import (
    open_phy,
    schedule_import_phy,
    import_phy_progress,
)
from antelope.utils.multithreading_utils import delete_thread_pool
from antelope.utils.os_utils import get_config
import pandas as pd
import numpy as np
import json


def show(username, tables):
    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:
        st.title("Electrophysiology")
        st.subheader("Manual Curation")

        st.divider()

        st.subheader("Select the data you want to manually curate")

        # get user to interactively select session
        _, spikesorting_dict = dropdown_query_table(
            tables,
            {"SpikeSorting": tables["SpikeSorting"]},
            username,
            headless=True,
            in_compute="False",
        )

        if spikesorting_dict == None:
            st.error("""You don't have any spikesorted data to manually curate.""")
            st.warning("""Please make sure you have spikesorted some data first.""")

            st.text("")

            # if there are any downlaods this session
            if "import_phy_jobs" in st.session_state:
                # button which shows spikesort statuses
                if st.button("Check upload progress"):
                    import_phy_progress()

        elif (
            "sortingparams_id" in spikesorting_dict.keys()
            and spikesorting_dict["experimenter"] == username
        ):
            st.divider()

            # schedule computation section
            st.subheader("Manual curation")

            config_dict = get_config()
            if config_dict["deployment"]["deployment"] == "apptainer":
                # ask for user password for scheduling the job
                password = st.text_input(
                    "Please enter your cluster password", type="password"
                )

            elif config_dict["deployment"]["deployment"] == "local":
                password = None

            st.text("")

            if st.button("Open phy"):
                st.text("")

                hashkey = (tables["SpikeSorting"] & spikesorting_dict).fetch1("phy")

                try:
                    open_phy(hashkey, username, password)

                except ImportError:
                    st.error("Phy not installed.")
                    st.warning(
                        "Please install phy by running: `pip install antelope[gui,phy]`"
                    )

                except Exception as e:
                    st.error("Error opening phy")
                    print(e)

                else:
                    st.warning(
                        """Please remember to save your manual curation results in phy when you're done!"""
                    )

            st.text("")

            st.info("This button will open phy in another window.")

            # upload data back to antelope

            st.divider()

            st.subheader("Upload manual curation results")

            # ask for user password for scheduling the job
            password = st.text_input(
                "Please enter your cluster password", type="password", key="2"
            )

            st.text("")

            if st.button("Upload results"):
                query = tables["SpikeSorting"] & spikesorting_dict
                in_compute = query.fetch1("spikesorting_in_compute")

                if in_compute == "True":
                    st.error(
                        """This entry is currently in a computation. You must wait until this job finishes to upload results."""
                    )

                else:
                    try:
                        # correction since datajoint returns np.int64 which isn't serialisable
                        if "labelledframes_in_compute" in spikesorting_dict.keys():
                            del spikesorting_dict["labelledframes_in_compute"]
                        del spikesorting_dict["spikesorting_in_compute"]
                        spikesorting_dict = {
                            key: (int(val) if isinstance(val, np.integer) else val)
                            for key, val in spikesorting_dict.items()
                        }
                        print(spikesorting_dict)

                        # send job
                        schedule_import_phy(
                            spikesorting_dict, username, password, numjobs=1
                        )

                        # success message
                        st.success("Upload job sent to cluster!")
                        st.info("You will receive an email once your job is completed")

                    except Exception as e:
                        print(e)

                        st.error("Error submitting job to cluster")

            st.text("")

            st.info(
                """This will schedule another job on the cluster that extracts the data from phy and reuploads it to the database.
            Note this will overwrite the existing electrophysiology results in the database.
            Please make sure you have saved your phy results before uploading!"""
            )

            st.text("")

            # if there are any downlaods this session
            if "import_phy_jobs" in st.session_state:
                # button which shows spikesort statuses
                if st.button("Check upload progress"):
                    import_phy_progress()

            # delete parameters section

            st.divider()

            st.subheader("Delete parameters")

            st.text("")

            if st.button("Check deletes"):
                # form delete query
                sort_dict = {
                    key: val
                    for key, val in spikesorting_dict.items()
                    if key
                    in [
                        "experimenter",
                        "experiment_id",
                        "animal_id",
                        "sortingparams_id",
                    ]
                }
                sort_dict["sortingparams_deleted"] = "False"
                query = tables["SortingParams"] & sort_dict
                # show cascaded deletes
                descendant_dict = show_deletes(tables, query)
                st.write("Entries that will get deleted:")
                st.dataframe(pd.DataFrame(data=descendant_dict).set_index("Table"))

            st.text("")

            st.info(
                "If your sorting parameters are not satisfactory, you can delete them here."
            )
            st.warning(
                "Note this will delete the sorting results for all sessions belonging to this parameter set."
            )

            st.text("")

            # input password and compare with hash on disk
            status = True  # change
            # status = bcrypt.checkpw(st.text_input(label='Please enter your antelope password to confirm deletion:',type='password').encode(),config['credentials']['usernames'][username]['password'].encode())

            st.text("")

            if st.button("Delete parameters"):
                # delete only if password correct
                if status:
                    # form delete query
                    sort_dict = {
                        key: int(val)
                        for key, val in spikesorting_dict.items()
                        if key in ["experiment_id", "animal_id", "sortingparams_id"]
                    }
                    sort_dict["experimenter"] = spikesorting_dict["experimenter"]
                    sort_dict["sortingparams_deleted"] = "False"
                    querystring = f"SortingParams & {json.dumps(sort_dict)}"
                    query = tables["SortingParams"] & sort_dict

                    # calculate number deletes
                    descendant_dict = show_deletes(tables, query)
                    num_delete = sum(descendant_dict["Number entries to be deleted"])

                    # if number deletes greater than 1000, use multithreading
                    if num_delete > 1000:
                        # retrieve thread pool
                        del_thread_pool = delete_thread_pool()

                        # submit job to thread pool
                        future_delete = del_thread_pool.submit(safe_delete, querystring)

                        # form query name
                        query_name = query_to_str(query)

                        # append future to session state
                        st.session_state.delete_futures.append(
                            (future_delete, query_name)
                        )

                        st.success("Deletion in progress.")

                    # otherwise delete in main thread
                    else:
                        safe_delete(querystring)

                        st.success("Deletion complete.")

                # incorrect password
                else:
                    st.error("Password incorrect.")



from antelope.connection import import_schemas, connect
from antelope.utils.os_utils import get_config
import getpass

config = get_config()
if config is None:
    print("Config file not found.")
    print("Please run `antelope-config` to generate a configuration file.")
    exit()

username = input("Please enter your username: ")
password = getpass.getpass("Please enter your password: ")

conn = connect.dbconnect(username, password)
tables = import_schemas.schema(conn)

# predefine experimenters
experimenters = [
    ["rbedford", "Rory Bedford", "Tripodi Lab", "MRC LMB", "True"],
    ["mtripodi", "Marcoo Tripodi", "Tripodi Lab", "MRC LMB", "True"],
    ["ewilliams", "Elena Williams", "Tripodi Lab", "MRC LMB", "False"],
    ["dwelch", "Daniel Welch", "Tripodi Lab", "MRC LMB", "False"],
    ["dmalmazet", "Daniel de Malmazet", "Tripodi Lab", "MRC LMB", "False"],
    ["srogers", "Stefan Rogers-Coltman", "Tripodi Lab", "MRC LMB", "False"],
    ["arueda", "Ana Gonzalez-Rueda", "Tripodi Lab", "MRC LMB", "False"],
    ["yyu", "Yujiao Yu", "Tripodi Lab", "MRC LMB", "False"],
    ["fmorgese", "Fabio Morgese", "Tripodi Lab", "MRC LMB", "False"],
]

# insert into the experimenter table
tables["Experimenter"].insert(experimenters, skip_duplicates=True)



import streamlit as st
import pandas as pd
from antelope.utils.datajoint_utils import (
    form_query,
    query_without_external,
    show_true_deletes,
    true_delete,
)
from antelope.utils.streamlit_utils import dropdown_query_table


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Database Administration")
        st.subheader("Permanently delete data")

        st.divider()

        # don't want to check experimenter table
        subtables = {key: val for key, val in tables.items() if key != "Experimenter"}

        # get input restriction
        tablename, restriction = dropdown_query_table(
            tables, subtables, username=None, delete_mode="True"
        )

        if not tablename:

            st.warning(
                "There is no temporarily deleted data currently in the database."
            )

        else:

            # form datajoint querystring
            querystring = f"{tablename} & {str(restriction)}"

            # form query
            query = form_query(querystring, tables)

            # make sure only searching deleted data
            query = query & {f"""{query.table_name.replace('_','')}_deleted""": "True"}

            st.text("")

            if st.button("Search"):

                st.text("")

                if querystring:

                    st.markdown("##### Deleted data:")

                    # query results to display
                    df, number = query_without_external(query)

                    # show table
                    if number == 0:
                        st.warning("No entries to show!")
                    else:
                        st.dataframe(df, hide_index=True)
                        st.text("Number of items: {}".format(number))
                else:
                    st.error("You must enter a query first!")

            st.text("")

            # warning message
            st.info(
                """
    Note that deletes cascade to avoid orphaned database entries.\n
    Please check your deletion before continuing."""
            )

            st.text("")

            if st.button("Check deletes"):

                # show cascaded deletes
                descendant_dict = show_true_deletes(tables, query)
                st.write("Entries that will get deleted:")
                st.dataframe(pd.DataFrame(data=descendant_dict).set_index("Table"))

            status = (
                st.text_input(
                    label="Please enter your password to confirm computation release:",
                    type="password",
                )
                == st.session_state.password
            )

            st.text("")

            # delete button
            if st.button("Delete", type="primary"):

                # delete only if password correct
                if status:

                    true_delete(query)

                    st.text("")

                    st.success("Data deleted!")

                # incorrect password
                else:
                    st.error("Password incorrect.")

            st.text("")

            # warning message
            st.error("Deletions are irreversible!")



import streamlit as st
from antelope.utils.streamlit_utils import dropdown_query_table
from antelope.utils.datajoint_utils import (
    query_without_external,
    form_query,
    query_to_str,
    show_restores,
    restore,
)
from antelope.utils.multithreading_utils import restore_thread_pool
import pandas as pd


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    with col2:

        st.title("Database Administration")
        st.subheader("Restore data")

        st.divider()

        # don't want to check experimenter table
        subtables = {key: val for key, val in tables.items() if key != "Experimenter"}

        # get input restriction
        tablename, restriction = dropdown_query_table(
            tables, subtables, username=None, delete_mode="True", restore=True
        )

        if not tablename:

            st.warning(
                "There is no temporarily deleted data currently in the database."
            )

        else:

            # form datajoint querystring
            querystring = f"{tablename} & {str(restriction)}"

            # form query
            query = form_query(querystring, tables)

            # make sure only searching deleted data
            query = query & {f"""{query.table_name.replace('_','')}_deleted""": "True"}

            st.text("")

            if st.button("Search"):

                st.text("")

                if querystring:

                    st.markdown("##### Deleted data:")

                    # query results to display
                    df, number = query_without_external(query)

                    # show table
                    if number == 0:
                        st.warning("No entries to show!")
                    else:
                        st.dataframe(df, hide_index=True, use_container_width=True)
                        st.text("Number of items: {}".format(number))
                else:
                    st.error("You must enter a query first!")

            st.text("")

            st.info(
                """Note that restores cascade to make it a full restore from the original delete. Please use the button below to check which tables will get data restored. If you want to restore only a subset of the data, the recommended workflow is to reverse the original delete, then re-delete the subset of the data you still don't want."""
            )

            st.text("")

            if st.button("Check restores"):

                # show cascaded deletes
                descendant_dict = show_restores(tables, query)
                st.write("Entries that will get restored:")
                st.dataframe(pd.DataFrame(data=descendant_dict).set_index("Table"))

            status = (
                st.text_input(
                    label="Please enter your password to confirm computation release:",
                    type="password",
                )
                == st.session_state.password
            )

            st.text("")

            # download data
            if st.button("Restore"):

                # calculate number of rows to restore
                descendant_dict = show_restores(tables, query)
                num_restores = sum(descendant_dict["Number entries to be restored"])

                # if number restores over 1000, run in separate thread
                if num_restores > 1000:

                    # retrieve thread pool
                    rest_thread_pool = restore_thread_pool()

                    # make name for query
                    query_name = query_to_str(form_query(querystring, tables))

                    # submit job to thread pool
                    future = rest_thread_pool.submit(restore, querystring)

                    # append future to session state
                    st.session_state.restore_futures.append((future, query_name))

                    st.text("")

                    st.success("Restore in progress!")

                # otherwise run in main thread
                else:

                    # restore data
                    restore(querystring)

                    # show success message
                    st.text("")
                    st.success("Data restored!")

            st.text("")

            # notice
            st.info(
                "Note that restoring data can take a while if there are a large number of rows to update. This will occur in a separate thread so you can still use Antelope while the restore is occurring, and can use the button below to check your download status."
            )

            st.text("")

            # add a button which shows download statuses
            # uses all downlaods in current session stored in session state
            if "restore_futures" in st.session_state:

                if st.button("Check restore progress"):

                    st.write("Restore statuses:")

                    # initialise data
                    display_futures = []

                    # display job progress
                    for future, query_name in st.session_state.restore_futures:

                        # compute statuses
                        if future.done():
                            if future.exception():
                                status = "restore error"
                            else:
                                status = "restore success"
                        else:
                            status = "restore in progress"

                        display_futures.append((query_name, status))

                    # make dataframe to display
                    df = pd.DataFrame(display_futures, columns=["Query", "Status"])

                    # show dataframe
                    st.dataframe(df, hide_index=True)



import toml
from pathlib import Path, PurePosixPath
from antelope.utils.os_utils import get_config_path


def run():

    print("\nWelcome to the antelope configuration setup tool.")

    config_path = get_config_path()
    if Path(config_path).exists():
        print(
            f"\nWarning, {config_path} already exists. Continuing will overwrite your current configuration."
        )
        cont = None
        while cont not in ["y", "n"]:
            cont = input("\nWould you like to continue? (y/n)\n")
            if cont == "n":
                exit()
            elif cont != "y":
                print("\nInvalid input. Please try again.\n")

    toml_dict = {"deployment": {"deployment": "local"}}

    toml_dict["mysql"] = {}
    toml_dict["mysql"]["host"] = input("\nPlease enter the MySQL host:\n")

    toml_dict["s3"] = {}
    toml_dict["s3"]["host"] = input("\nPlease enter the S3 host:\n")

    toml_dict["folders"] = {}
    status = True
    print(
        "\nYou can add as many folders as you like which will be made searchable within Antelope."
    )
    print(
        "For example, you can add your home directory, plus a mounted cluster directory."
    )
    print("Please give each an identifiable name and path.")

    while status == True:

        name = input("\nPlease enter the folder name:\n")
        path = input("\nPlease enter the (absolute) folder path:\n")

        try:
            assert Path(path).is_absolute(), "\nInvalid path. Please try again."
        except AssertionError as e:
            print(e)
            continue

        toml_dict["folders"][name] = path
        cont = None

        while cont not in ["y", "n"]:
            cont = input("\nWould you like to add another folder? (y/n)\n")
            if cont == "n":
                status = False
            elif cont != "y":
                print("\nInvalid input. Please try again.\n")

    toml_dict["analysis"] = {}
    status = True
    print(
        "\nYou can add folders that contain your custom analysis scripts written to work with antelope."
    )
    print("Antelope will automatically load all functions found in these folders.")

    toml_dict["analysis"]["folders"] = []

    cont = None
    while cont not in ["y", "n"]:
        cont = input("\nWould you like to add a custom analysis path? (y/n)\n")
        if cont == "n":
            status = False
        elif cont != "y":
            print("\nInvalid input. Please try again.\n")

    while status == True:

        path = input("\nPlease enter the (absolute) folder path:\n")

        try:
            assert Path(path).is_absolute(), "\nInvalid path. Please try again."
        except AssertionError as e:
            print(e)
            continue

        toml_dict["analysis"]["folders"].append(path)
        cont = None

        while cont not in ["y", "n"]:
            cont = input("\nWould you like to add another folder? (y/n)\n")
            if cont == "n":
                status = False
            elif cont != "y":
                print("\nInvalid input. Please try again.\n")

    toml_dict["multithreading"] = {}
    print("\nAntelope can use multiple cores in the background to speed up processing.")
    status = False
    while not status:
        try:
            cores = input("Please enter the maximum number of cores to use:\n")
            assert cores.isdigit(), "\nThe number of cores must be an integer."
        except AssertionError as e:
            print(e)
            print("Please try again.\n")
        else:
            toml_dict["multithreading"]["max_workers"] = int(cores)
            status = True

    toml_dict["computation"] = {}
    toml_dict["computation"]["host"] = input(
        "\nPlease enter the host name of the compute cluster:\n"
    )

    status = False
    while not status:
        try:
            basedir = input(
                "\nPlease enter the base directory for your antelope installation on the compute cluster:\n"
            )
            assert PurePosixPath(
                basedir
            ).is_absolute(), "\nInvalid path. Please try again."
        except AssertionError as e:
            print(e)
            print("Please try again.\n")
        else:
            toml_dict["computation"]["basedir"] = basedir
            status = True

    status = False
    while not status:
        try:
            antelope_data = input(
                "\nPlease enter the path to your antelope data directory on the compute cluster as mounted on this machine:\n"
            )
            assert Path(
                antelope_data
            ).is_absolute(), "\nInvalid path. Please try again."
        except AssertionError as e:
            print(e)
            print("Please try again.\n")
        else:
            toml_dict["computation"]["antelope_data"] = antelope_data
            status = True

    print(f"\nWriting configuration to {config_path}")

    config_path.parent.mkdir(parents=True, exist_ok=True)

    with open(config_path, "w") as f:
        toml.dump(toml_dict, f)



import streamlit as st
import streamlit_antd_components as sac
from antelope.utils import analysis_utils
from antelope.utils.streamlit_utils import display_analysis, server_directory_browser
from antelope.utils.external_utils import schedule_analysis, analysis_progress
import json


def show(username, tables):

    col1, col2, col3 = st.columns([0.2, 0.6, 0.2])

    if not hasattr(st.session_state, "result"):
        st.session_state.result = None

    with col2:

        st.title("Analysis")
        st.subheader("Rerun an analysis")

        st.markdown(
            """###### Rerun an analysis from a saved reproducibility json file, checking the code and data haven't changed."""
        )

        st.divider()

        repr_json = server_directory_browser(
            "Select your reproducibility json", extension="json"
        )

        st.divider()

        if repr_json:
            with open(repr_json, "r") as f:
                repr_dict = json.load(f)

            # get the function from its name
            functions = analysis_utils.import_analysis(
                tables["Experimenter"].connection, tables
            )
            folder = functions[repr_dict["name"].split(".")[0]]
            func_dict = {f.name: f for f in folder}
            for fct in folder:
                if fct.name == repr_dict["name"].split(".")[1]:
                    function = fct
                    break

            if isinstance(function.query, str):
                table = tables[function.query].proj()
            elif isinstance(function.query, list):
                table = tables[function.query[0]]
                for q in function.query[1:]:
                    table = table * tables[q].proj()

        if st.button("Check hashes"):
            if repr_json:
                # check the hashes
                message = function.check_hash(repr_json)

                # print messages
                if message == "Reproducibility checks passed.":
                    st.text("")
                    st.success("Reproducibility checks passed.")
                else:
                    st.text("")
                    st.warning(message)

            else:
                st.text("")
                st.error("Please select a reproducibility json file.")

        # run the function
        st.divider()
        st.markdown("#### Select mode")
        sac.buttons(
            [sac.ButtonsItem(label="Local"), sac.ButtonsItem(label="Cluster")],
            key="mode",
            use_container_width=True,
        )

        if st.session_state.mode == "Local":
            st.text("")
            result = None
            if st.button("Run"):

                if repr_json:
                    # get the restriction
                    restriction = repr_dict["restriction"]
                    # get the args
                    args = repr_dict["arguments"]

                    @st.cache_data(ttl=600)
                    def cache_fct(name, restriction, **args):
                        fct = func_dict[name]
                        return fct(restriction, **args)

                    result = cache_fct(function.name, restriction, **args)
                    st.session_state.result = result

                    st.divider()
                    display_analysis(
                        st.session_state.result, function.returns, table.primary_key
                    )

                else:
                    st.error("Please select a reproducibility json file.")

        if st.session_state.mode == "Cluster":
            if repr_json:
                st.text("")

                savepath = st.text_input("Enter the path to save the results to")
                numcpus = st.number_input(
                    "Enter the number of cpus to use",
                    min_value=1,
                    max_value=64,
                    value=1,
                )
                time = st.number_input(
                    "Enter the time to run the job for (minutes)",
                    min_value=10,
                    max_value=24 * 60,
                    value=60,
                )
                password = st.text_input(
                    "Please enter your cluster password", type="password"
                )

                st.text("")
                if st.button("Schedule analysis job"):

                    try:
                        schedule_analysis(
                            function,
                            restriction,
                            savepath,
                            numcpus,
                            time,
                            password,
                            args,
                        )
                    except Exception:
                        st.error(
                            "Job scheduling failed. Please check your inputs and try again."
                        )
                    else:
                        st.text("")
                        st.success("Job sent to cluster.")
            else:
                st.error("Please select a reproducibility json file.")

            # if there are any downlaods this session
            if "analysis_jobs" in st.session_state:

                # button which shows spikesort statuses
                st.text("")
                if st.button("Check analysis progress"):

                    analysis_progress()



from antelope.connection import connect, import_schemas
from antelope.utils.os_utils import get_config
from antelope.utils.analysis_utils import import_analysis
import os
import types

config = get_config()
username = os.environ.get("DB_USER")
password = os.environ.get("DB_PASS")

if username is None or password is None:
    raise Exception("Please set the DB_USER and DB_PASS environment variables")

conn = connect.dbconnect(username, password)
tables = import_schemas.schema(conn)

for key, val in tables.items():
    globals()[key] = val

# import analysis functions
analysis_functions = import_analysis(conn, tables)
for folder, fcts in analysis_functions.items():
    # make class
    if folder not in globals():
        globals()[folder] = types.SimpleNamespace()
    for fct in fcts:
        setattr(globals()[folder], fct.name, fct)



"""
Utility functions involving applications outside the main antelope container.

Author: Rory Bedford
Date: 02/11/23
"""

import json
import paramiko
import os
import socket
from pathlib import Path, PurePosixPath
import math
import streamlit as st
from datetime import datetime, timedelta
from antelope.utils.os_utils import get_config
import numpy as np


def slurm_time(numjobs, scale=1):
    """
    Function calculates how long to schedule master job for
    """
    # function is a sigmoid starting at around two hours, and saturating at a day at around 100 params
    t = scale * 3600 * (24 / (1 + math.exp((-numjobs + 50) / 20)))

    # convert to slurm format
    t_delta = timedelta(seconds=t)
    timestring = str(t_delta).split(".")[0]
    if len(timestring) == 7:
        timestring = "0" + timestring

    return timestring


def compute_log():
    """
    Computes directory in which to store log files for this job
    """
    # current time
    d = datetime.now()

    # logdir
    logdir = PurePosixPath(str(d.year)) / str(d.month) / str(d.day)

    return str(logdir)


def schedule_spikesorting(session_dict, username, password, numjobs):
    """
    Function calls the nextflow spikesorting pipeline
    """

    config = get_config()

    # get hostname
    if config["deployment"]["deployment"] == "apptainer":
        node = socket.gethostname()
    elif config["deployment"]["deployment"] == "local":
        node = config["computation"]["host"]

    # path for bash submission
    compute_dict = config["computation"]
    basedir = PurePosixPath(compute_dict["basedir"])
    script = str(basedir / "workflows" / "submit.sh")

    # load credentials for nextflow environment
    credentials = dict(
        DB_HOST=config["mysql"]["host"],
        DB_USER=st.session_state.username,
        DB_PASS=st.session_state.password,
        S3_HOST=config["s3"]["host"],
        S3_USER=st.session_state.username,
        S3_PASS=st.session_state.password,
    )

    # make slurm env var string
    env_string = ",".join([f"{key}='{val}'" for key, val in credentials.items()])

    # make nextflow submission command
    param = json.dumps(session_dict, separators=(",", ":"))

    logdir = compute_log()
    logdirmnt = Path(config["computation"]["antelope_data"]) / "logs" / logdir

    slurmtime = slurm_time(numjobs)

    # command to be run
    command = (
        f"""bash {script} '{param}' {logdir} spikesorting {env_string} {slurmtime}"""
    )

    # initialise paramiko
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    # connect to node
    ssh.connect(
        node,
        username=username,
        password=password,
        look_for_keys=False,
        allow_agent=False,
    )

    # execute command
    stdin, stdout, stderr = ssh.exec_command(command)
    bash_id = stdout.read().decode().strip()

    # initialise session state
    if "spikesort_jobs" not in st.session_state:
        st.session_state.spikesort_jobs = []

    # append job id and logdir to session state
    st.session_state.spikesort_jobs.append((bash_id, logdirmnt))


def spikesorting_progress():
    """
    Function checks progress of currently running spikesorting jobs
    """
    for i, val in enumerate(st.session_state.spikesort_jobs):

        st.divider()
        st.text("")

        st.write(f"Job {i+1} progress:")

        # read log file
        slurmfile = Path(os.path.expandvars(val[1])) / f"slurm-{val[0]}.out"
        try:
            with open(slurmfile, "r") as f:
                log = f.read()

            # split log at double newline
            if len(log.split("\n\nexecutor")) > 1:
                log = log.split("\n\nexecutor")[-1]
                st.text(log)
            else:
                st.text("\nJob awaiting cluster submission\n")

        except FileNotFoundError:
            st.text("\nJob awaiting cluster submission\n")


def schedule_import_phy(ephys_dict, username, password, numjobs):
    """
    Function calls the nextflow spikesorting pipeline
    """

    config = get_config()

    # get hostname
    if config["deployment"]["deployment"] == "apptainer":
        node = socket.gethostname()
    elif config["deployment"]["deployment"] == "local":
        node = config["computation"]["host"]

    # path for bash submission
    compute_dict = config["computation"]
    basedir = PurePosixPath(compute_dict["basedir"])
    script = str(basedir / "workflows" / "submit.sh")

    # load credentials for nextflow environment
    credentials = dict(
        DB_HOST=config["mysql"]["host"],
        DB_USER=st.session_state.username,
        DB_PASS=st.session_state.password,
        S3_HOST=config["s3"]["host"],
        S3_USER=st.session_state.username,
        S3_PASS=st.session_state.password,
    )

    # make slurm env var string
    env_string = ",".join([f"{key}='{val}'" for key, val in credentials.items()])

    # make nextflow submission command
    param = json.dumps(ephys_dict, separators=(",", ":"))

    logdir = compute_log()
    logdirmnt = Path(config["computation"]["antelope_data"]) / "logs" / logdir

    slurmtime = slurm_time(numjobs)

    # command to be run
    command = (
        f"""bash {script} '{param}' {logdir} import_phy {env_string} {slurmtime}"""
    )

    # initialise paramiko
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    # execute command
    ssh.connect(
        node,
        username=username,
        password=password,
        look_for_keys=False,
        allow_agent=False,
    )
    stdin, stdout, stderr = ssh.exec_command(command)
    bash_id = stdout.read().decode().strip()

    # initialise session state
    if "import_phy_jobs" not in st.session_state:
        st.session_state.import_phy_jobs = []

    # append job id and logdir to session state
    st.session_state.import_phy_jobs.append((bash_id, logdirmnt))


def import_phy_progress():
    """
    Function checks progress of currently running spikesorting jobs
    """
    for i, val in enumerate(st.session_state.import_phy_jobs):

        st.divider()
        st.text("")

        st.write(f"Job {i+1} progress:")

        # read log file
        slurmfile = Path(os.path.expandvars(val[1])) / f"slurm-{val[0]}.out"
        try:
            with open(slurmfile, "r") as f:
                log = f.read()

            # split log at double newline
            if len(log.split("\n\nexecutor")) > 1:
                log = log.split("\n\nexecutor")[-1]
                st.text(log)
            else:
                st.text("\nJob awaiting cluster submission\n")

        except FileNotFoundError:
            st.text("\nJob awaiting cluster submission\n")


def open_phy(hashkey, username, password):
    """
    Function opens phy on the given ephys entry
    """

    config_dict = get_config()

    if config_dict["deployment"]["deployment"] == "apptainer":
        # need to change this since we don't configure cluster data location anymore

        # get hostname
        node = socket.gethostname()

        # calculate phy container
        basedir = Path(config_dict["computation"]["basedir"])
        phycontainer = str(basedir / "containers" / "phy.sif")

        # where phy data get stored
        phybase = Path(config_dict["computation"]["antelope_data"]) / "phy"
        phypath = str(phybase / hashkey / "phy")

        # command to be run
        command = f"""DISPLAY={os.environ['DISPLAY']} apptainer run -B {phypath}:/mnt {phycontainer}"""

        # initialise paramiko
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # execute command
        ssh.connect(
            node,
            username=username,
            password=password,
            look_for_keys=False,
            allow_agent=False,
        )
        stdin, stdout, stderr = ssh.exec_command(command)

    elif config_dict["deployment"]["deployment"] == "local":

        from phy.apps.template import template_gui
        from antelope.utils.multithreading_utils import phy_thread_pool

        phybase = Path(config_dict["computation"]["antelope_data"]) / "phy"
        phypath = str(phybase / hashkey / "phy" / "params.py")

        phy_thread = phy_thread_pool()
        phy_thread.submit(template_gui, phypath)


def schedule_train_dlc(key, num_videos, password):
    """
    Function calls the nextflow spikesorting pipeline
    """

    config = get_config()

    # get hostname
    if config["deployment"]["deployment"] == "apptainer":
        node = socket.gethostname()
    elif config["deployment"]["deployment"] == "local":
        node = config["computation"]["host"]

    # path for bash submission
    compute_dict = config["computation"]
    basedir = PurePosixPath(compute_dict["basedir"])
    script = str(basedir / "workflows" / "submit.sh")

    # load credentials for nextflow environment
    credentials = dict(
        DB_HOST=config["mysql"]["host"],
        DB_USER=st.session_state.username,
        DB_PASS=st.session_state.password,
        S3_HOST=config["s3"]["host"],
        S3_USER=st.session_state.username,
        S3_PASS=st.session_state.password,
    )

    # make slurm env var string
    env_string = ",".join([f"{key}='{val}'" for key, val in credentials.items()])

    # make key json serialisable
    key = {
        k: (int(v) if k in ["experiment_id", "behaviourrig_id", "dlcmodel_id"] else v)
        for k, v in key.items()
    }
    del key["labelledframes_deleted"]

    # make nextflow submission command
    param = json.dumps(key, separators=(",", ":"))

    logdir = compute_log()
    logdirmnt = Path(config["computation"]["antelope_data"]) / "logs" / logdir

    # convert to slurm format
    slurmtime = slurm_time(num_videos * 0.2, 5)

    # command to be run
    command = f"""bash {script} '{param}' {logdir} train_dlc {env_string} {slurmtime}"""

    # initialise paramiko
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    # execute command
    ssh.connect(
        node,
        username=st.session_state.username,
        password=password,
        look_for_keys=False,
        allow_agent=False,
    )
    stdin, stdout, stderr = ssh.exec_command(command)
    bash_id = stdout.read().decode().strip()

    # initialise session state
    if "train_dlc_jobs" not in st.session_state:
        st.session_state.train_dlc_jobs = []

    # append job id and logdir to session state
    st.session_state.train_dlc_jobs.append((bash_id, logdirmnt))


def train_dlc_progress():
    """
    Function checks progress of currently running dlc training jobs
    """
    for i, val in enumerate(st.session_state.train_dlc_jobs):

        st.divider()
        st.text("")

        st.write(f"Job {i+1} progress:")

        # read log file
        slurmfile = Path(os.path.expandvars(val[1])) / f"slurm-{val[0]}.out"
        try:
            with open(slurmfile, "r") as f:
                log = f.read()

            # split log at double newline
            if len(log.split("\n\nexecutor")) > 1:
                log = log.split("\n\nexecutor")[-1]
                st.text(log)
            else:
                st.text("\nJob awaiting cluster submission\n")

        except FileNotFoundError:
            st.text("\nJob awaiting cluster submission\n")


def schedule_analysis(function, restriction, savepath, numcpus, time, password, args):
    """
    Function sends analysis job to cluster.
    """

    for key, val in restriction.items():
        if isinstance(val, np.int64):
            restriction[key] = int(val)

    # get function name and script
    func_name = function.name
    func_folder = function.folder

    # get other arguments
    restriction_str = json.dumps(restriction, separators=(",", ":"))
    args_str = json.dumps(args, separators=(",", ":"))

    # submission path
    config = get_config()
    node = config["computation"]["host"]
    compute_dict = config["computation"]
    basedir = PurePosixPath(compute_dict["basedir"])
    script = str(basedir / "workflows" / "analysis" / "submit.sh")

    # convert time to slurm format
    t_delta = timedelta(seconds=time * 60)
    timestring = str(t_delta).split(".")[0]
    if len(timestring) == 7:
        timestring = "0" + timestring

    # make slurm env var string
    credentials = dict(
        DB_HOST=config["mysql"]["host"],
        DB_USER=st.session_state.username,
        DB_PASS=st.session_state.password,
        S3_HOST=config["s3"]["host"],
        S3_USER=st.session_state.username,
        S3_PASS=st.session_state.password,
    )
    env_string = ",".join([f"{key}='{val}'" for key, val in credentials.items()])

    logdir = compute_log()
    logdirmnt = Path(config["computation"]["antelope_data"]) / "logs" / logdir

    # command to be run
    command = f"""bash {script} {logdir} {env_string} {timestring} {numcpus} {savepath} '{func_folder}' '{func_name}' '{restriction_str}' '{args_str}'"""

    # initialise paramiko
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    # connect to cluster
    ssh.connect(
        node,
        username=st.session_state.username,
        password=password,
        look_for_keys=False,
        allow_agent=False,
    )

    # execute command
    stdin, stdout, stderr = ssh.exec_command(command)
    bash_id = stdout.read().decode().strip()

    # initialise session state
    if "analysis_jobs" not in st.session_state:
        st.session_state.analysis_jobs = []

    # append job id and logdir to session state
    st.session_state.analysis_jobs.append((bash_id, logdirmnt))


def analysis_progress():
    """
    Function checks progress of currently running analysis jobs
    """
    for i, val in enumerate(st.session_state.analysis_jobs):

        st.divider()
        st.text("")

        st.write(f"Job {i+1} progress:")

        # read log file
        slurmfile = Path(os.path.expandvars(val[1])) / f"slurm-{val[0]}.out"
        try:
            with open(slurmfile, "r") as f:
                log = f.read()

            st.text(log)

        except FileNotFoundError:
            st.text("\nJob awaiting cluster submission\n")


def schedule_inference_dlc(key, num_videos, password):
    """
    Function calls the nextflow spikesorting pipeline
    """

    config = get_config()

    # get hostname
    if config["deployment"]["deployment"] == "apptainer":
        node = socket.gethostname()
    elif config["deployment"]["deployment"] == "local":
        node = config["computation"]["host"]

    # path for bash submission
    compute_dict = config["computation"]
    basedir = PurePosixPath(compute_dict["basedir"])
    script = str(basedir / "workflows" / "submit.sh")

    # load credentials for nextflow environment
    credentials = dict(
        DB_HOST=config["mysql"]["host"],
        DB_USER=st.session_state.username,
        DB_PASS=st.session_state.password,
        S3_HOST=config["s3"]["host"],
        S3_USER=st.session_state.username,
        S3_PASS=st.session_state.password,
    )

    # make slurm env var string
    env_string = ",".join([f"{key}='{val}'" for key, val in credentials.items()])

    # make key json serialisable
    key = {
        k: (
            int(v)
            if k in ["experiment_id", "behaviourrig_id", "dlcmodel_id", "session_id"]
            else v
        )
        for k, v in key.items()
    }

    # make nextflow submission command
    param = json.dumps(key, separators=(",", ":"))

    logdir = compute_log()
    logdirmnt = Path(config["computation"]["antelope_data"]) / "logs" / logdir

    # convert to slurm format
    slurmtime = slurm_time(num_videos * 0.2, 5)

    # command to be run
    command = (
        f"""bash {script} '{param}' {logdir} inference_dlc {env_string} {slurmtime}"""
    )

    # initialise paramiko
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    # execute command
    ssh.connect(
        node,
        username=st.session_state.username,
        password=password,
        look_for_keys=False,
        allow_agent=False,
    )
    stdin, stdout, stderr = ssh.exec_command(command)
    bash_id = stdout.read().decode().strip()

    # initialise session state
    if "inference_dlc_jobs" not in st.session_state:
        st.session_state.inference_dlc_jobs = []

    # append job id and logdir to session state
    st.session_state.inference_dlc_jobs.append((bash_id, logdirmnt))


def inference_dlc_progress():
    """
    Function checks progress of currently running dlc inferenceing jobs
    """
    for i, val in enumerate(st.session_state.inference_dlc_jobs):

        st.divider()
        st.text("")

        st.write(f"Job {i+1} progress:")

        # read log file
        slurmfile = Path(os.path.expandvars(val[1])) / f"slurm-{val[0]}.out"
        try:
            with open(slurmfile, "r") as f:
                log = f.read()

            # split log at double newline
            if len(log.split("\n\nexecutor")) > 1:
                log = log.split("\n\nexecutor")[-1]
                st.text(log)
            else:
                st.text("\nJob awaiting cluster submission\n")

        except FileNotFoundError:
            st.text("\nJob awaiting cluster submission\n")



from antelope.utils.analysis_base import antelope_analysis
from antelope.utils.analysis_utils import split_trials
from antelope.utils.antelope_utils import (
    insert_nwb,
    upload_rig_json,
    recompute_masks,
    export_nwb,
)



"""
This module contains the standard library spike-triggered average analysis functions.
"""

import numpy as np
from antelope import antelope_analysis
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt


@antelope_analysis
class AnalogSta:
    """
    The spike-triggered average for an analog event.
    """

    name = "analog_sta"
    query = ["SpikeTrain", "AnalogEvents"]
    returns = {"Spike-triggered average": np.ndarray, "Timestamps (s)": np.ndarray}
    args = {"window_size": float, "sample_rate": float}

    def run(key, window_size=1, sample_rate=1000):

        spiketrain = (SpikeTrain & key).fetch1("spiketrain")
        data, timestamps = (AnalogEvents.proj("data", "timestamps") & key).fetch1(
            "data", "timestamps"
        )

        # interpolate the event data
        event_func = interp1d(timestamps, data, fill_value=0, bounds_error=False)

        # create window timestamps
        step = 1 / sample_rate
        start_time = -(window_size // step) * step
        window_timestamps = np.arange(start_time, 0, step)

        # create matrix of window times for each spike - shape (n_spikes, window_samples)
        sta_times = spiketrain[:, None] + window_timestamps

        # get the event values in each window
        sta_values = event_func(sta_times)

        # average over all spikes
        sta = np.mean(sta_values, axis=0)

        return sta, window_timestamps


@antelope_analysis
class PlotAnalogSta:
    """
    Plot the spike-triggered average for an analog event.
    """

    name = "plot_analog_sta"
    query = ["SpikeTrain", "AnalogEvents"]
    returns = {"Spike-triggered average": plt.Figure}
    args = {"window_size": float, "sample_rate": float}

    def run(key, window_size=1, sample_rate=1000):

        unit, name = (AnalogEvents.proj("unit", "analogevents_name") & key).fetch1(
            "unit", "analogevents_name"
        )

        result = analog_sta(key, window_size, sample_rate)
        sta, timestamps = result["Spike-triggered average"], result["Timestamps (s)"]

        fig, ax = plt.subplots()

        ax.plot(timestamps, sta)
        ax.set_xlabel("Time (s)")
        ax.set_ylabel(f"{name} ({unit})")
        ax.set_title("Spike-triggered average")

        return fig


@antelope_analysis
class DigitalSta:
    """
    The spike-triggered average for a digital event.
    """

    name = "digital_sta"
    query = ["SpikeTrain", "DigitalEvents"]
    returns = {"Spike-triggered average": np.ndarray, "Timestamps (s)": np.ndarray}
    args = {"window_size": float, "sample_rate": float}

    def run(key, window_size=1, sample_rate=1000):

        spiketrain = (SpikeTrain & key).fetch1("spiketrain")
        data, timestamps = (DigitalEvents.proj("data", "timestamps") & key).fetch1(
            "data", "timestamps"
        )

        if spiketrain.size > 0:

            if timestamps.size == 0:
                start_time = spiketrain[0] - window_size
                end_time = spiketrain[-1]
            else:
                start_time = min(timestamps[0], spiketrain[0] - window_size)
                end_time = max(timestamps[-1], spiketrain[-1])

            global_timestamps = np.arange(start_time, end_time, 1 / sample_rate)

            # get the indices of each spike in the global timestamps array
            spiketrain_indices = np.digitize(spiketrain, global_timestamps) - 1

            # make event data match global timestamps, filled with zeros
            event_indices = np.digitize(timestamps, global_timestamps) - 1
            event_data = np.zeros_like(global_timestamps)
            event_data[event_indices] = data

            # create window array - shape (n_spikes, window_samples)
            window_indices = np.arange(-window_size * sample_rate + 1, 0, 1)
            window_array = spiketrain_indices[:, None] + window_indices
            window_timestamps = window_indices / sample_rate

            # get the event values in each window
            sta_values = event_data[window_array]

            # average over all spikes
            sta = np.mean(sta_values, axis=0)

        else:
            sta = np.array([])
            window_timestamps = np.array([])

        return sta, window_timestamps


@antelope_analysis
class PlotDigitalSta:
    """
    Plot the spike-triggered average for an analog event.
    """

    name = "plot_digital_sta"
    query = ["SpikeTrain", "DigitalEvents"]
    returns = {"Spike-triggered average": plt.Figure}
    args = {"window_size": float, "sample_rate": float}

    def run(key, window_size=1, sample_rate=1000):

        unit, name = (DigitalEvents.proj("unit", "digitalevents_name") & key).fetch1(
            "unit", "digitalevents_name"
        )

        result = digital_sta(key, window_size, sample_rate)
        sta, timestamps = result["Spike-triggered average"], result["Timestamps (s)"]

        fig, ax = plt.subplots()

        ax.plot(timestamps, sta)
        ax.set_xlabel("Time (s)")
        ax.set_ylabel(f"{name} ({unit})")
        ax.set_title("Spike-triggered average")

        return fig


@antelope_analysis
class IntervalSta:
    """
    The spike-triggered average for a digital event.
    """

    name = "interval_sta"
    query = ["SpikeTrain", "IntervalEvents"]
    returns = {"Spike-triggered average": np.ndarray, "Timestamps (s)": np.ndarray}
    args = {"window_size": float, "sample_rate": float}

    def run(key, window_size=1, sample_rate=1000):

        spiketrain = (SpikeTrain & key).fetch1("spiketrain")
        data, timestamps = (IntervalEvents.proj("data", "timestamps") & key).fetch1(
            "data", "timestamps"
        )

        # delete this, just since some test data corrupted
        if np.any(data == 0):
            return np.array([]), np.array([])

        if timestamps.size == 0:
            window_timestamps = np.arange(-window_size, 0, 1 / sample_rate)
            sta = np.zeros_like(window_timestamps)

        else:
            start_time = min(timestamps[0], spiketrain[0] - window_size)
            end_time = max(timestamps[-1], spiketrain[-1])

            global_timestamps = np.arange(start_time, end_time, 1 / sample_rate)

            # get the indices of each spike in the global timestamps array
            spiketrain_indices = np.digitize(spiketrain, global_timestamps) - 1

            # make event data match global timestamps
            event_indices = np.digitize(global_timestamps, timestamps) - 1
            event_data = data[event_indices]
            event_data[event_data == -1] = 0
            event_data[event_indices == -1] = 0

            # create window array - shape (n_spikes, window_samples)
            window_indices = np.arange(-window_size * sample_rate + 1, 0, 1)
            window_array = spiketrain_indices[:, None] + window_indices
            window_timestamps = window_indices / sample_rate

            # get the event values in each window
            sta_values = event_data[window_array]

            # average over all spikes
            sta = np.mean(sta_values, axis=0)

        return sta, window_timestamps


@antelope_analysis
class PlotIntervalSta:
    """
    Plot the spike-triggered average for an interval event.
    """

    name = "plot_interval_sta"
    query = ["SpikeTrain", "IntervalEvents"]
    returns = {"Spike-triggered average": plt.Figure}
    args = {"window_size": float, "sample_rate": float}

    def run(key, window_size=1, sample_rate=1000):

        name = (IntervalEvents.proj("intervalevents_name") & key).fetch1(
            "intervalevents_name"
        )

        result = interval_sta(key, window_size, sample_rate)
        sta, timestamps = result["Spike-triggered average"], result["Timestamps (s)"]

        fig, ax = plt.subplots()

        ax.plot(timestamps, sta)
        ax.set_xlabel("Time (s)")
        ax.set_ylabel(f"{name}")
        ax.set_title("Spike-triggered average")

        return fig



from pathlib import Path, PurePosixPath
import platform
import toml
import shutil
import os


def get_config_path():
    if os.environ.get("ANTELOPE_CONFIG_PATH") is None:
        if platform.system() in ["Linux", "Darwin", "Windows"]:
            config_path = Path.home() / ".config" / "antelope" / "config.toml"

            return config_path

    else:
        config_path = Path(os.environ.get("ANTELOPE_CONFIG_PATH"))
        return config_path


def get_config():
    config_path = get_config_path()
    if not config_path.exists():
        return None
    else:
        with open(config_path, "r") as f:
            config = toml.load(f)
        return config


def validate_config_file(config):
    keys = {
        "deployment",
        "mysql",
        "s3",
        "multithreading",
        "computation",
        "folders",
        "analysis",
    }
    valid = keys.issubset(config.keys())
    return valid


def cp_st_config():
    if platform.system() in ["Linux", "Darwin", "Windows"]:
        # copy  streamlit config to home if it doesn's exist
        stconfig = Path.home() / ".streamlit" / "config.toml"
        stcredentials = Path.home() / ".streamlit" / "credentials.toml"
        if not stconfig.exists():
            stconfig.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(
                Path(os.path.abspath(__file__)).parent.parent
                / "configs"
                / ".streamlit"
                / "config.toml",
                stconfig,
            )
        if not stcredentials.exists():
            stcredentials.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(
                Path(os.path.abspath(__file__)).parent.parent
                / "configs"
                / ".streamlit"
                / "credentials.toml",
                stcredentials,
            )


def validate_config(config):
    for name, path in config["folders"].items():
        if not Path(path).exists():
            return False
    for folder in config["analysis"]["folders"]:
        if not Path(folder).exists():
            return False
    cluster_install = PurePosixPath(config["computation"]["basedir"])
    if not cluster_install.is_absolute():
        return False
    cluster_data = Path(config["computation"]["antelope_data"])
    if not cluster_data.exists():
        return False
    return True



import streamlit_antd_components as sac


def show(username, tables):

    sac.alert(
        label="Permission denied",
        description="You do not have admin privileges",
        color="red",
        icon=True,
    )



import datajoint as dj


def schema(conn):

    schema = dj.Schema(schema_name="antelope_ephys", connection=conn)

    # import the metadata schema
    metadata = dj.create_virtual_module("metadata", "antelope_metadata")

    @schema
    class ProbeGeometry(dj.Manual):
        definition = """
        # Probeinterface ProbeGroup files used across animals
        -> metadata.Experimenter
        probegeometry_id : smallint # Unique probe ID (auto_increment)
        ---
        probe : json # ProbeInterface format
        probegeometry_name : varchar(40) # Short probe description
        probegeometry_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class SortingParams(dj.Manual):
        definition = """
        # Parameters to be passed to the spike sorting pipeline
        -> metadata.Animal
        sortingparams_id : smallint # Unique params ID (auto_increment)
        ---
        sortingparams_name : varchar(40) # Short sortingparams description
        manually_sorted = 'False' : enum('False', 'True') # Data externally spike sorted
        params : json # Spikesorting parameters
        sortingparams_notes: varchar(1000) # Optional sorting parameters description
        sortingparams_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class ProbeInsertion(dj.Manual):
        definition = """
        # Probe insertion for this animal
        -> metadata.Animal
        ---
        -> ProbeGeometry
        yaw : decimal(3, 0) # Probe extrinsic rotation relative to dv axis (deg)
		pitch : decimal(3, 0) # Probe extrinsic rotation relative to ap axis (deg)
		roll : decimal(3, 0) # Probe extrinsic rotation relative to ml axis (deg)
		ap_coord : decimal(5, 0) # Probe anterior-posterior coordinate relative to bregma (um)
		ml_coord : decimal(5, 0) # Probe medial-lateral coordinate relative to bregma (um)
		dv_coord : decimal(5, 0) # Probe dorsal-ventral coordinate relative to bregma (um)
        probeinsertion_notes : varchar(1000) # Optional probe insertion description
        probeinsertion_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function

        """

    @schema
    class Recording(dj.Manual):
        definition = """
        # Recording for this animal
        -> metadata.Animal
        -> metadata.Session
        ---
        recording : attach@raw_ephys # Recording folder
		ephys_acquisition: enum('axona','openephys') # Equipment type
        probe_dv_increment : decimal(4, 0) # Probe dorsal-ventral coordinate increment relative to previous session (um)
        recording_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class SpikeSorting(dj.Imported):
        definition = """
        # Parent table for all curated and populated ephys data
        -> metadata.Session
        -> SortingParams
        ---
        phy : varchar(200) # Tracks phy folders for manual curation
        manually_curated : enum('False','True') # Has the data been manually curated
        spikesorting_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function
        spikesorting_in_compute : enum('False','True') # Implements row locking
        """

    @schema
    class Probe(dj.Computed):
        definition = """
        # Usually a tetrode but can be any valid probe (such as a neuropixel probe)
        -> SpikeSorting
        probe_id : int # Given by probegeometry file
        ---
        probe_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class Channel(dj.Computed):
        definition = """
        # Corresponds to a single electrode on the probe
        -> Probe
        channel_id : int # Given by probegeometry file
        ---
        ap_coord : decimal(5, 0) # Probe anterior-posterior coordinate relative to bregma (mm)
        ml_coord : decimal(5, 0) # Probe medial-lateral coordinate relative to bregma (mm)
        dv_coord : decimal(5, 0) # Probe dorsal-ventral coordinate relative to bregma (mm)
        channel_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class LFP(dj.Computed):
        definition = """
        # Local field potential
        -> Channel
        ---
        lfp : longblob # LFP array for session, low-pass filtered, in uV
        lfp_sample_rate : int # Set to be 2.5 times the sample rate
        lfp_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class Unit(dj.Computed):
        definition = """
        # Unit found by spikesorting
        -> Probe
        unit_id : int # Unique ID for this unit
        ---
        ap_coord : decimal(5, 0) # Probe anterior-posterior coordinate relative to bregma (mm)
        ml_coord : decimal(5, 0) # Probe medial-lateral coordinate relative to bregma (mm)
        dv_coord : decimal(5, 0) # Probe dorsal-ventral coordinate relative to bregma (mm)
        unit_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class SpikeTrain(dj.Computed):
        definition = """
        # Timestamps for when the unit fires
        -> Unit
        ---
        spiketrain: mediumblob # Numpy array of spike times in seconds
        spiketrain_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function
        """

    @schema
    class Waveform(dj.Computed):
        definition = """
        # Waveform for each time the unit fires
        -> Unit
        -> Channel
        ---
        waveform : longblob # Numpy array shape n*m, where n is number of spikes, m is number of samples, in uV
        waveform_sample_rate : int # Original sample rate from acquisition system
        ms_before : float # Milliseconds before peak extracted
        ms_after : float # Milliseconds after peak extracted
        waveform_deleted = 'False' : enum('False', 'True') # Implements a temporary delete function
        """

    tables = {
        "ProbeGeometry": ProbeGeometry,
        "SortingParams": SortingParams,
        "ProbeInsertion": ProbeInsertion,
        "Recording": Recording,
        "SpikeSorting": SpikeSorting,
        "Probe": Probe,
        "Channel": Channel,
        "LFP": LFP,
        "Unit": Unit,
        "SpikeTrain": SpikeTrain,
        "Waveform": Waveform,
    }

    return tables, schema
